<?php

namespace VSHM\Modules;

use VSHM\Bus\ChangeReservationCustomer;
use VSHM\Bus\ChangeReservationDate;
use VSHM\Bus\ChangeReservationService;
use VSHM\Bus\ChangeReservationStatus;
use VSHM\Bus\CreateReservation;
use VSHM\Bus\DeleteReservation;
use VSHM\Bus\DeleteReservationProperty;
use VSHM\Bus\UpdateOrCreateReservationProperty;
use VSHM\Functions;
use VSHM\Plugin\DateTimeTbk;
use VSHM\Providers\Customers;
use VSHM\Providers\Reservations;
use VSHM\Providers\ReservationsData;
use VSHM\Providers\Services;
use VSHM\Providers\ServicesData;
use VSHM\REST_Controller;
use VSHM\Settings\Reservation\ZoomMeetingId;
use VSHM\Settings\Service\ConfirmationEmailToAdmin;
use VSHM\Settings\Service\CreateZoomMeeting;
use VSHM\Settings\Service\Personal_ConfirmationEmailToProvider;
use VSHM\Settings\ZoomApiKey;
use VSHM\Settings\ZoomApiSecret;
use VSHM\Tools;

defined('ABSPATH') || exit;

/**
 * Class Zoom
 *
 * @author VonStroheim
 */
class Zoom
{
    /**
     * @var string
     */
    public static $route_path = '/zoom';

    public static function bootstrap(): void
    {
        REST_Controller::register_routes([
            self::$route_path . '/reservation/getmeetinglink'    => [
                'methods'  => \WP_REST_Server::READABLE,
                'callback' => [self::class, 'getMeetingLinkCallback'],
                'args'     => [
                    'hash'          => [
                        'type' => 'string'
                    ],
                    'reservationId' => [
                        'type'     => 'string',
                        'required' => TRUE
                    ],
                ]
            ],
            self::$route_path . '/reservation/startmeetinglink/' => [
                'methods'             => \WP_REST_Server::READABLE,
                'callback'            => [self::class, 'startMeetingLinkCallback'],
                'args'                => [
                    'reservationId' => [
                        'type'     => 'string',
                        'required' => TRUE
                    ],
                ],
                'permission_callback' => function () {
                    return Functions::current_user_is_provider();
                }
            ]
        ]);
        add_filter('tbk_backend_settings_page_items', [self::class, 'add_settings_page_items']);
        if (isset(vshm()->settings()[ ZoomApiSecret::ID ], vshm()->settings()[ ZoomApiKey::ID ])
            && vshm()->settings()[ ZoomApiSecret::ID ]
            && vshm()->settings()[ ZoomApiKey::ID ]
        ) {
            add_action('vshm_dispatched_CreateReservation', [self::class, 'create_meeting_after_res']);
            add_action('vshm_dispatching_DeleteReservation', [self::class, 'delete_meeting_after_removing_res']);
            add_action('vshm_dispatched_ChangeReservationStatus', [self::class, 'change_reservation_status_callback'], 10, 3);
            add_action('vshm_dispatched_ChangeReservationService', [self::class, 'change_reservation_service_callback'], 10, 3);
            add_action('vshm_dispatched_ChangeReservationCustomer', [self::class, 'change_reservation_customer_callback'], 10, 3);
            add_action('vshm_dispatched_ChangeReservationDate', [self::class, 'change_reservation_date_callback'], 10, 3);
            add_filter('tbk_preparing_reservation_for_frontend', [self::class, 'inject_meeting_id']);
            add_filter('tbk_reservation_data_processing_frontend', [self::class, 'reservation_data_processing']);
            add_filter('tbk_template_hooks_editor', [self::class, 'template_hooks_editor'], 10, 3);
            add_filter('tbk_notification_templates', [self::class, 'template_hooks'], 10, 3);
        }
    }

    /**
     * @param array       $values
     * @param string      $reservationId
     * @param string|null $notificationType
     *
     * @return array
     */
    public static function template_hooks(array $values, string $reservationId, ?string $notificationType = ''): array
    {
        $meetingId = ReservationsData::provideBy(['reservation_id' => $reservationId, 'key' => ZoomMeetingId::ID], TRUE);
        if ($meetingId) {
            $meeting              = self::getMeeting($meetingId);
            $values['zoom::join'] = $meeting['join_url'];
            if ($notificationType === ConfirmationEmailToAdmin::ID || $notificationType === Personal_ConfirmationEmailToProvider::ID) {
                $values['zoom::start'] = $meeting['start_url'];
            }
        }

        return $values;
    }

    /**
     * @param array  $hooks
     * @param string $notificationType
     * @param string $serviceId
     *
     * @return array
     */
    public static function template_hooks_editor(array $hooks, string $notificationType, string $serviceId): array
    {
        $hooks[] = [
            'value'        => 'zoom::join',
            'label'        => __('Zoom Join Meeting link', 'team-booking'),
            'context'      => 'reservation',
            'contextLabel' => __('Reservation', 'team-booking')
        ];
        if ($notificationType === ConfirmationEmailToAdmin::ID || $notificationType === Personal_ConfirmationEmailToProvider::ID) {
            $hooks[] = [
                'value'        => 'zoom::start',
                'label'        => __('Zoom Start Meeting link', 'team-booking'),
                'context'      => 'reservation',
                'contextLabel' => __('Reservation', 'team-booking')
            ];
        }


        return $hooks;
    }

    /**
     * @param array $datum
     *
     * @return array
     */
    public static function reservation_data_processing(array $datum): array
    {
        if ($datum['key'] === ZoomMeetingId::ID) {
            $meeting       = self::getMeeting($datum['value']);
            $mappedMeeting = [
                'id' => $datum['value']
            ];
            if (isset($meeting['join_url'])) {
                $mappedMeeting['join_url'] = $meeting['join_url'];
            }
            if (isset($meeting['start_url'])) {
                $mappedMeeting['start_url'] = $meeting['start_url'];
            }
            if (isset($meeting['password'])) {
                $mappedMeeting['password'] = $meeting['password'];
            }
            if (isset($meeting['status'])) {
                $mappedMeeting['status'] = $meeting['status']; //"waiting" or "started"
            }
            $datum['value'] = $mappedMeeting;
        }

        return $datum;
    }

    /**
     * @param array $reservation
     *
     * @return array
     */
    public static function inject_meeting_id(array $reservation): array
    {
        $meetingIds = array_column(ReservationsData::provideBy(['key' => ZoomMeetingId::ID]), 'value', 'reservation_id');
        if (isset($meetingIds[ $reservation['id'] ])) {
            $reservation[ ZoomMeetingId::ID ] = $meetingIds[ $reservation['id'] ];
        }

        return $reservation;
    }

    /**
     * @param ChangeReservationDate     $command
     * @param string                    $agent_type
     * @param                           $agent_id
     *
     * @return void
     */
    public static function change_reservation_date_callback(ChangeReservationDate $command, string $agent_type, $agent_id): void
    {
        $reservation = Reservations::provideBy(['id' => $command->getId()], TRUE);
        if (!$reservation) {
            return;
        }

        $meetingId = ReservationsData::provideBy(['reservation_id' => $reservation['id'], 'key' => ZoomMeetingId::ID], TRUE);

        if ($meetingId) {
            $startDate = DateTimeTbk::createFromFormatSilently('U', $reservation['start']);
            self::updateMeeting($meetingId, [
                'start_time' => $startDate->format('Y-m-d\TH:i:s\Z'),
                'duration'   => round(($reservation['end'] - $reservation['start']) / 60),
            ]);
        }
    }

    /**
     * @param ChangeReservationCustomer $command
     * @param string                    $agent_type
     * @param                           $agent_id
     *
     * @return void
     */
    public static function change_reservation_customer_callback(ChangeReservationCustomer $command, string $agent_type, $agent_id): void
    {
        $reservation = Reservations::provideBy(['id' => $command->getId()], TRUE);
        if (!$reservation) {
            return;
        }

        $meetingId = ReservationsData::provideBy(['reservation_id' => $reservation['id'], 'key' => ZoomMeetingId::ID], TRUE);

        if ($meetingId) {
            $customer = Customers::provideBy(['id' => $reservation['customerId']], TRUE);
            self::updateMeeting($meetingId, [
                'agenda' => sprintf(
                /* translators: %s: Name of a customer */
                    __('Meeting with %s', 'team-booking'),
                    $customer['name']
                ),
            ]);
        }
    }

    /**
     * @param ChangeReservationService $command
     * @param string                   $agent_type
     * @param                          $agent_id
     *
     * @return void
     */
    public static function change_reservation_service_callback(ChangeReservationService $command, string $agent_type, $agent_id): void
    {
        $reservation = Reservations::provideBy(['id' => $command->getId()], TRUE);
        if (!$reservation) {
            return;
        }
        $meetingId = ReservationsData::provideBy(['reservation_id' => $reservation['id'], 'key' => ZoomMeetingId::ID], TRUE);

        if ($meetingId) {
            self::deleteMeeting($meetingId);
            vshm()->bus->dispatch(new DeleteReservationProperty($reservation['id'], ZoomMeetingId::ID));
        }
        self::createMeeting($reservation);
    }

    /**
     * @param ChangeReservationStatus $command
     * @param string                  $agent_type
     * @param                         $agent_id
     *
     * @return void
     */
    public static function change_reservation_status_callback(ChangeReservationStatus $command, string $agent_type, $agent_id): void
    {
        if ($agent_type === \VSHM\Bus\Bus::AGENT_USER) {
            // When AGENT_USER, we are facing "admin" changes of the status not in consequence of a specific action.
            return;
        }
        $reservation = Reservations::provideBy(['id' => $command->getId()], TRUE);
        if (!$reservation) {
            return;
        }
        $meetingId = ReservationsData::provideBy(['reservation_id' => $reservation['id'], 'key' => ZoomMeetingId::ID], TRUE);
        switch ($reservation['status']) {
            case 'confirmed':
                if (!$meetingId) {
                    self::createMeeting($reservation);
                }
                break;
            case 'cancelled':
                if ($meetingId) {
                    self::deleteMeeting($meetingId);
                    vshm()->bus->dispatch(new DeleteReservationProperty($reservation['id'], ZoomMeetingId::ID));
                }
                break;
            default:
                break;
        }
    }

    /**
     * @param CreateReservation $command
     *
     * @return void
     */
    public static function create_meeting_after_res(CreateReservation $command): void
    {
        $reservation = Reservations::provideBy(['id' => $command->getId()], TRUE);
        if ($reservation && $reservation['status'] === 'confirmed') {
            self::createMeeting($reservation);
        }
    }

    /**
     * @param DeleteReservation $command
     *
     * @return void
     */
    public static function delete_meeting_after_removing_res(DeleteReservation $command): void
    {
        $meetingId = ReservationsData::provideBy(['reservation_id' => $command->getId(), 'key' => ZoomMeetingId::ID], TRUE);
        if ($meetingId) {
            self::deleteMeeting($meetingId);
        }
    }

    /**
     * @param \WP_REST_Request $request
     *
     * @return \WP_REST_Response
     */
    public static function getMeetingLinkCallback(\WP_REST_Request $request): \WP_REST_Response
    {
        $reservation = Reservations::provideBy(['id' => $request->get_param('reservationId')], TRUE);
        $meetingId   = ReservationsData::provideBy(['reservation_id' => $request->get_param('reservationId'), 'key' => ZoomMeetingId::ID], TRUE);
        $response    = [];
        if ($reservation && $meetingId) {
            $customer = Customers::provideBy(['id' => $reservation['customerId']], TRUE);

            if ($customer['wp_user'] === get_current_user_id() || md5($reservation['id'] . $customer['access_token']) === $request->get_param('hash')) {
                $meeting = self::getMeeting($meetingId);
                if (!$meeting) {
                    return REST_Controller::get_error_response(REST_Controller::get_root_rest_url() . self::$route_path . '/reservation/getmeetinglink', [
                        'message' => 'Zoom API error. See the error.log for details.'
                    ]);
                }
                if (isset($meeting['join_url'])) {
                    $response['link'] = $meeting['join_url'];
                }
            }
        }

        return REST_Controller::get_ok_response(REST_Controller::get_root_rest_url() . self::$route_path . '/reservation/getmeetinglink', $response);
    }

    /**
     * @param \WP_REST_Request $request
     *
     * @return \WP_REST_Response
     */
    public static function startMeetingLinkCallback(\WP_REST_Request $request): \WP_REST_Response
    {
        $reservation = Reservations::provideBy(['id' => $request->get_param('reservationId')], TRUE);
        $meetingId   = ReservationsData::provideBy(['reservation_id' => $request->get_param('reservationId'), 'key' => ZoomMeetingId::ID], TRUE);
        $response    = [];
        if ($reservation && $meetingId) {
            $meeting = self::getMeeting($meetingId);
            if (!$meeting) {
                return REST_Controller::get_error_response(REST_Controller::get_root_rest_url() . self::$route_path . '/reservation/startmeetinglink', [
                    'message' => 'Zoom API error. See the error.log for details.'
                ]);
            }
            if (isset($meeting['start_url'])) {
                $response['link'] = $meeting['start_url'];
            }
        }

        return REST_Controller::get_ok_response(REST_Controller::get_root_rest_url() . self::$route_path . '/reservation/startmeetinglink', $response);

    }

    public static function getMeeting($id)
    {
        $zoom = self::getZoom();

        $response = $zoom->doRequest('GET', '/meetings/{meetingId}', [], ['meetingId' => $id]);

        if ($response === FALSE) {
            error_log("Errors:" . implode("\n", $zoom->requestErrors()));
        }

        return $response;
    }

    public static function getMeetings()
    {
        $zoom = self::getZoom();

        $response = $zoom->doRequest('GET', '/users/me/meetings');
        if ($response === FALSE) {
            error_log("Errors:" . implode("\n", $zoom->requestErrors()));
        }

        return $response;
    }

    public static function deleteMeeting($id)
    {
        $zoom     = self::getZoom();
        $response = $zoom->doRequest('DELETE', '/meetings/{meetingId}', [], ['meetingId' => $id]);

        if ($response === FALSE) {
            error_log("Errors:" . implode("\n", $zoom->requestErrors()));
        }

        return $response;
    }

    public static function updateMeeting($meetingId, $updated = [])
    {
        $zoom     = self::getZoom();
        $response = $zoom->doRequest('PATCH', '/meetings/{meetingId}', [], ['meetingId' => $meetingId], json_encode($updated, JSON_FORCE_OBJECT));
        if ($response === FALSE) {
            error_log("Errors:" . implode("\n", $zoom->requestErrors()));
        }
        if ($zoom->responseCode() !== 204) {
            error_log("Errors:" . implode("\n", $response['message']));
        }
    }

    public static function createMeeting($reservation): bool
    {
        $service    = Services::provideBy(['id' => $reservation['serviceId']], TRUE);
        $createZoom = filter_var(ServicesData::provideBy(['service_id' => $reservation['serviceId'], 'key' => CreateZoomMeeting::ID], TRUE), FILTER_VALIDATE_BOOLEAN);

        if ($service && $createZoom) {

            $customer = Customers::provideBy(['id' => $reservation['customerId']], TRUE);

            $startDate = DateTimeTbk::createFromFormatSilently('U', $reservation['start']);
            $startDate->setTimezone(new \DateTimeZone('UTC'));
            $endDate = DateTimeTbk::createFromFormatSilently('U', $reservation['end']);
            $endDate->setTimezone(new \DateTimeZone('UTC'));

            $zoom = self::getZoom();

            // Gathering password requirements...

            $response = $zoom->doRequest('GET', '/users/me/settings', ['option' => 'meeting_security']);

            if ($zoom->responseCode() !== 200) {
                ob_start();
                echo 'Zoom API error: ';
                var_dump($response);
                error_log(ob_get_clean());

                return FALSE;
            }

            $pwd_req  = $response['meeting_security']['meeting_password_requirement'];
            $pwd_type = 'alnum';
            $pwd_len  = 10;
            if ($pwd_req['length']) {
                $pwd_len = (int)$pwd_req['length'];
            }
            if ($pwd_req['only_allow_numeric']) {
                $pwd_type = 'numeric';
            }
            if ($pwd_req['have_special_character']) {
                $pwd_len--;
            }
            if ($pwd_req['have_upper_and_lower_characters']) {
                $pwd_len -= 2;
            }
            $pwd = Tools::generate_token($pwd_type, $pwd_len);
            if ($pwd_req['have_special_character']) {
                $pwd .= '@';
            }
            if ($pwd_req['have_upper_and_lower_characters']) {
                $pwd .= 'zZ';
            }

            $meeting = [
                'topic'      => $service['name'],
                'type'       => 2,
                'start_time' => $startDate->format('Y-m-d\TH:i:s\Z'),
                'duration'   => round(($endDate->getTimestamp() - $startDate->getTimestamp()) / 60),
                'password'   => $pwd,
                'agenda'     => sprintf(
                /* translators: %s: Name of a customer */
                    __('Meeting with %s', 'team-booking'),
                    $customer['name']
                ),
                'settings'   => [

                ]
            ];

            $response = $zoom->doRequest('POST', '/users/me/meetings', [], [], json_encode($meeting, JSON_FORCE_OBJECT));
            if ($response === FALSE) {
                error_log("Errors:" . implode("\n", $zoom->requestErrors()));
            }

            if ($zoom->responseCode() !== 201) {
                ob_start();
                var_dump($response);
                error_log(ob_get_clean());
            } else if (isset($response['id'])) {
                vshm()->bus->dispatch(new UpdateOrCreateReservationProperty($reservation['id'], ZoomMeetingId::ID, $response['id']));
            }
        }

        return TRUE;
    }

    /**
     * @return ZoomAPIWrapper
     */
    public static function getZoom(): ZoomAPIWrapper
    {
        return new ZoomAPIWrapper(vshm()->settings()[ \VSHM\Settings\ZoomApiKey::ID ], vshm()->settings()[ \VSHM\Settings\ZoomApiSecret::ID ]);
    }

    public static function add_settings_page_items($page)
    {
        $menu_item      = \VSHM\UI\Admin\SidebarItem::option(__('Zoom', 'team-booking'), 'zoom');
        $settings_panel = \VSHM\UI\Admin\SettingsPanel::get();


        $settings_panel->setTitle(__('Zoom', 'team-booking'));
        $settings_panel->setIconUrl(vshm()->plugin['URL'] . '/Assets/zoom_logo.png');
        $settings_panel->setDescription(__('How to get your API keys? You need to go to the Zoom "App Marketplace", log in using your zoom account and then select "Build App" from the "Develop" dropdown menu. Then you need to choose the JWT option for "server-to-server integration". Once you will out some basic information you will be taken to the "App Credentials" page which will give you your API key and API secret.', 'team-booking'));

        $settings_panel->addItem(\VSHM\Settings\ZoomApiKey::getBackendElement());
        $settings_panel->addItem(\VSHM\Settings\ZoomApiSecret::getBackendElement());

        $menu_item->setContent($settings_panel);
        $page->addSidebarItem($menu_item);

        return $page;
    }
}

/*
==============================================================================
MIT License
Copyright (c) 2020 Ben Jefferson
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
==============================================================================
This is a simple wrapper class to handle authenticating requests to the Zoom APIs
Usage:
$zoom = new ZoomAPIWrapper( '<your API key>', '<your API secret>' );
// It is up to you to use the right method, path and specify the path parameters
// to match the {placeholders} in the path.
// You can find all the details of method, path, placholders and body content in the Zoom
// API reference docs here: https://marketplace.zoom.us/docs/api-reference/zoom-api
// $queryParameters,$requestParameters,$bodyContent are all optional
$response = $zoom->doRequest($method,$path,$queryParameters,$requestParameters,$bodyContent);
if ($response === false) {
    // There was an error before the request was event sent to the api
    echo "Errors:".implode("\n",$zoom->requestErrors())
} esle {
    printf("Response code: %d\n",$zoom->responseCode());
    print_r($response);
}
*/

class ZoomAPIWrapper
{
    private $errors;
    private $apiKey;
    private $apiSecret;
    private $baseUrl;
    private $timeout;

    public function __construct($apiKey, $apiSecret, $options = array())
    {
        $this->apiKey = $apiKey;

        $this->apiSecret = $apiSecret;

        $this->baseUrl = 'https://api.zoom.us/v2';
        $this->timeout = 30;

        // Store any options if they map to valid properties
        foreach ($options as $key => $value) {
            if (property_exists($this, $key)) $this->$key = $value;
        }
    }

    public static function urlsafeB64Encode($string)
    {
        return str_replace('=', '', strtr(base64_encode($string), '+/', '-_'));
    }

    private function generateJWT()
    {
        $token  = array(
            'iss' => $this->apiKey,
            'exp' => time() + 60,
        );
        $header = array(
            'typ' => 'JWT',
            'alg' => 'HS256',
        );

        $toSign =
            self::urlsafeB64Encode(json_encode($header))
            . '.' .
            self::urlsafeB64Encode(json_encode($token));

        $signature = hash_hmac('SHA256', $toSign, $this->apiSecret, TRUE);

        return $toSign . '.' . self::urlsafeB64Encode($signature);
    }

    private function headers()
    {
        return array(
            'Authorization: Bearer ' . $this->generateJWT(),
            'Content-Type: application/json',
            'Accept: application/json',
        );
    }

    private function pathReplace($path, $requestParams)
    {
        $errors = array();
        $path   = preg_replace_callback('/\\{(.*?)\\}/', static function ($matches) use ($requestParams, &$errors) {
            if (!isset($requestParams[ $matches[1] ])) {
                $errors[] = 'Required path parameter was not specified: ' . $matches[1];

                return '';
            }

            return rawurlencode($requestParams[ $matches[1] ]);
        }, $path);

        if (count($errors)) $this->errors = array_merge($this->errors, $errors);

        return $path;
    }

    public function doRequest($method, $path, $queryParams = array(), $pathParams = array(), $body = '')
    {

        if (is_array($body)) {
            // Treat an empty array in the body data as if no body data was set
            if (!count($body)) $body = '';
            else $body = json_encode($body);
        }

        $this->errors       = array();
        $this->responseCode = 0;

        $path = $this->pathReplace($path, $pathParams);

        if (count($this->errors)) return FALSE;

        $method = strtoupper($method);
        $url    = $this->baseUrl . $path;

        // Add on any query parameters
        if (count($queryParams)) $url .= '?' . http_build_query($queryParams);

        $ch = curl_init();

        /**
         * Enable this only for LOCAL test
         * TODO: REMOVE IN PROD
         */
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);

        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $this->headers());
        curl_setopt($ch, CURLOPT_TIMEOUT, $this->timeout);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);

        if (in_array($method, array('DELETE', 'PATCH', 'POST', 'PUT'))) {

            // All except DELETE can have a payload in the body
            if ($method != 'DELETE' && strlen($body)) {
                curl_setopt($ch, CURLOPT_POST, TRUE);
                curl_setopt($ch, CURLOPT_POSTFIELDS, $body);
            }

            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
        }

        $result = curl_exec($ch);

        // Check the return value of curl_exec()
        if ($result === FALSE) {
            throw new \RuntimeException(curl_error($ch), curl_errno($ch));
        }

        $contentType        = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        $this->responseCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);

        curl_close($ch);

        return json_decode($result, TRUE);
    }

    // Returns the errors responseCode returned from the last call to doRequest
    public function requestErrors()
    {
        return $this->errors;
    }

    // Returns the responseCode returned from the last call to doRequest
    public function responseCode()
    {
        return $this->responseCode;
    }
}