<?php

namespace VSHM;

defined('ABSPATH') || exit;

if (!class_exists(Session_Controller::class)) {

    /**
     * Class Session_Controller
     *
     * @author  VonStroheim
     * @package VSHM
     */
    class Session_Controller
    {
        /**
         * Customer ID.
         *
         * @var int $_customer_id Customer ID.
         */
        protected $_customer_id;

        /**
         * Session Data.
         *
         * @var array $_data Data array.
         */
        protected $_data = [];

        /**
         * Dirty when the session needs saving.
         *
         * @var bool $_dirty When something changes
         */
        protected $_dirty = FALSE;

        /**
         * Cookie name used for the session.
         *
         * @var string cookie name
         */
        protected $_cookie;

        /**
         * Stores session expiry.
         *
         * @var string session due to expire timestamp
         */
        protected $_session_expiring;

        /**
         * Stores session due to expire timestamp.
         *
         * @var string session expiration timestamp
         */
        protected $_session_expiration;

        /**
         * True when the cookie exists.
         *
         * @var bool Based on whether a cookie exists.
         */
        protected $_has_cookie = FALSE;

        /**
         * Table name for session data.
         *
         * @var string Custom session table name
         */
        protected $_table;

        public function __construct()
        {
            $this->_cookie = apply_filters('tbk_cookie', 'wp_tbk_session_' . COOKIEHASH);
            $this->_table  = $GLOBALS['wpdb']->prefix . 'tbk_sessions';
            DB::create_table('tbk_sessions', [
                'session_key'    => ['type' => 'char', 'chars' => 32, 'primary' => TRUE],
                'session_value'  => 'text',
                'session_expiry' => 'int',
            ]);
        }

        public function init()
        {
            $this->init_session_cookie();

            add_action('tbk_set_cart_cookies', [$this, 'set_customer_session_cookie'], 10);
            add_action('shutdown', [$this, 'save_data'], 20);
            add_action('wp_logout', [$this, 'destroy_session']);

            if (!is_user_logged_in()) {
                add_filter('nonce_user_logged_out', [$this, 'maybe_update_nonce_user_logged_out'], 10, 2);
            }
        }

        /**
         * Get customer ID.
         *
         * This is NOT the reservation customer ID
         *
         * @return int
         */
        public function get_customer_id()
        {
            return $this->_customer_id;
        }

        /**
         * Magic get method.
         *
         * @param mixed $key Key to get.
         *
         * @return mixed
         */
        public function __get($key)
        {
            return $this->get($key);
        }

        /**
         * Magic set method.
         *
         * @param mixed $key   Key to set.
         * @param mixed $value Value to set.
         */
        public function __set($key, $value)
        {
            $this->set($key, $value);
        }

        /**
         * Magic isset method.
         *
         * @param mixed $key Key to check.
         *
         * @return bool
         */
        public function __isset($key)
        {
            return isset($this->_data[ sanitize_title($key) ]);
        }

        /**
         * Magic unset method.
         *
         * @param mixed $key Key to unset.
         */
        public function __unset($key)
        {
            if (isset($this->_data[ $key ])) {
                unset($this->_data[ $key ]);
                $this->_dirty = TRUE;
            }
        }

        /**
         * Get a session variable.
         *
         * @param string $key     Key to get.
         * @param mixed  $default used if the session variable isn't set.
         *
         * @return array|string value of session variable
         */
        public function get($key, $default = NULL)
        {
            $key = sanitize_key($key);

            return isset($this->_data[ $key ]) ? maybe_unserialize($this->_data[ $key ]) : $default;
        }

        /**
         * Set a session variable.
         *
         * @param string $key   Key to set.
         * @param mixed  $value Value to set.
         */
        public function set($key, $value)
        {
            if ($value !== $this->get($key)) {
                $this->_data[ sanitize_key($key) ] = maybe_serialize($value);
                $this->_dirty                      = TRUE;
            }
        }

        /**
         * Setup cookie and customer ID.
         */
        public function init_session_cookie()
        {
            $cookie = $this->get_session_cookie();

            if ($cookie) {
                // Customer ID will be an MD5 hash if this is a guest session.
                $this->_customer_id        = $cookie[0];
                $this->_session_expiration = $cookie[1];
                $this->_session_expiring   = $cookie[2];
                $this->_has_cookie         = TRUE;
                $this->_data               = $this->get_session_data();

                if (!$this->is_session_cookie_valid()) {
                    $this->destroy_session();
                    $this->set_session_expiration();
                }

                // If the user logs in, update session.
                if (is_user_logged_in() && (string)get_current_user_id() !== $this->_customer_id) {
                    $guest_session_id   = $this->_customer_id;
                    $this->_customer_id = (string)get_current_user_id();
                    $this->_dirty       = TRUE;
                    $this->save_data($guest_session_id);
                    $this->set_customer_session_cookie(TRUE);
                }

                // Update session if it is close to expiring.
                if (time() > $this->_session_expiring) {
                    $this->set_session_expiration();
                    $this->update_session_timestamp($this->_customer_id, $this->_session_expiration);
                }
            } else {
                $this->set_session_expiration();
                $this->_customer_id = $this->generate_customer_id();
                $this->_data        = $this->get_session_data();
            }
        }

        /**
         * Checks if session cookie is expired, or belongs to a logged-out user.
         *
         * @return bool Whether session cookie is valid.
         */
        private function is_session_cookie_valid()
        {
            // If session is expired, session cookie is invalid.
            if (time() > $this->_session_expiration) {
                return FALSE;
            }

            // If user has logged out, session cookie is invalid.
            if (!is_user_logged_in() && !$this->is_customer_guest($this->_customer_id)) {
                return FALSE;
            }

            // Session from a different user is not valid. (Although from a guest user will be valid)
            if (is_user_logged_in() && !$this->is_customer_guest($this->_customer_id) && (string)get_current_user_id() !== $this->_customer_id) {
                return FALSE;
            }

            return TRUE;
        }

        /**
         * Sets the session cookie on-demand (usually after adding an item to the cart).
         *
         * Warning: Cookies will only be set if this is called before the headers are sent.
         *
         * @param bool $set Should the session cookie be set.
         */
        public function set_customer_session_cookie($set)
        {
            if ($set) {
                $to_hash           = $this->_customer_id . '|' . $this->_session_expiration;
                $cookie_hash       = hash_hmac('md5', $to_hash, wp_hash($to_hash));
                $cookie_value      = $this->_customer_id . '||' . $this->_session_expiration . '||' . $this->_session_expiring . '||' . $cookie_hash;
                $this->_has_cookie = TRUE;

                if (!isset($_COOKIE[ $this->_cookie ]) || $_COOKIE[ $this->_cookie ] !== $cookie_value) {
                    tbk_setcookie($this->_cookie, $cookie_value, $this->_session_expiration, $this->use_secure_cookie(), TRUE);
                }
            }
        }

        /**
         * Should the session cookie be secure?
         *
         * @return bool
         */
        protected function use_secure_cookie()
        {
            return apply_filters('tbk_session_use_secure_cookie', strpos(get_option('home'), 'https:') !== FALSE && is_ssl());
        }

        /**
         * Return true if the current user has an active session, i.e. a cookie to retrieve values.
         *
         * @return bool
         */
        public function has_session()
        {
            return isset($_COOKIE[ $this->_cookie ]) || $this->_has_cookie || is_user_logged_in();
        }

        /**
         * Set session expiration.
         */
        public function set_session_expiration()
        {
            $this->_session_expiring   = time() + (int)apply_filters('tbk_session_expiring', 60 * 60 * 47); // 47 Hours.
            $this->_session_expiration = time() + (int)apply_filters('tbk_session_expiration', 60 * 60 * 48); // 48 Hours.
        }

        /**
         * Generate a unique customer ID for guests, or return user ID if logged in.
         *
         * This is NOT the customer ID that is generated for the reservations!
         *
         * Uses Portable PHP password hashing framework to generate a unique cryptographically strong ID.
         *
         * @return string
         */
        public function generate_customer_id()
        {
            $customer_id = '';

            if (is_user_logged_in()) {
                $customer_id = (string)get_current_user_id();
            }

            if (empty($customer_id)) {
                require_once ABSPATH . 'wp-includes/class-phpass.php';
                $hasher      = new \PasswordHash(8, FALSE);
                $customer_id = 'tbk_' . substr(md5($hasher->get_random_bytes(32)), 2);
            }

            return $customer_id;
        }

        /**
         * Checks if this is an auto-generated customer ID.
         *
         * @param string|int $customer_id Customer ID to check.
         *
         * @return bool Whether customer ID is randomly generated.
         */
        private function is_customer_guest($customer_id)
        {
            $customer_id = (string)$customer_id;

            if (empty($customer_id)) {
                return TRUE;
            }

            if (strpos($customer_id, 'tbk_') === 0) {
                return TRUE;
            }

            return FALSE;
        }

        /**
         * Get the session cookie, if set. Otherwise, return false.
         *
         * Session cookies without a customer ID are invalid.
         *
         * @return bool|array
         */
        public function get_session_cookie()
        {
            $cookie_value = isset($_COOKIE[ $this->_cookie ]) ? wp_unslash($_COOKIE[ $this->_cookie ]) : FALSE;

            if (empty($cookie_value) || !is_string($cookie_value)) {
                return FALSE;
            }

            [$customer_id, $session_expiration, $session_expiring, $cookie_hash] = explode('||', $cookie_value);

            if (empty($customer_id)) {
                return FALSE;
            }

            // Validate hash.
            $to_hash = $customer_id . '|' . $session_expiration;
            $hash    = hash_hmac('md5', $to_hash, wp_hash($to_hash));

            if (empty($cookie_hash) || !hash_equals($hash, $cookie_hash)) {
                return FALSE;
            }

            return [$customer_id, $session_expiration, $session_expiring, $cookie_hash];
        }

        /**
         * Get session data.
         *
         * @return array
         */
        public function get_session_data()
        {
            return $this->has_session() ? (array)$this->get_session($this->_customer_id, []) : [];
        }

        /**
         * Save data and delete guest session.
         *
         * @param int $old_session_key session ID before user logs in.
         */
        public function save_data($old_session_key = 0)
        {
            // Dirty if something changed - prevents saving nothing new.

            if ($this->_dirty && $this->has_session()) {
                global $wpdb;

                $wpdb->query(
                    $wpdb->prepare(
                        "INSERT INTO $this->_table (`session_key`, `session_value`, `session_expiry`) VALUES (%s, %s, %d)
 					ON DUPLICATE KEY UPDATE `session_value` = VALUES(`session_value`), `session_expiry` = VALUES(`session_expiry`)",
                        $this->_customer_id,
                        maybe_serialize($this->_data),
                        $this->_session_expiration
                    )
                );

                $this->_dirty = FALSE;
                if (get_current_user_id() !== $old_session_key && !is_object(get_user_by('id', $old_session_key))) {
                    $this->delete_session($old_session_key);
                }
            }
        }

        /**
         * Destroy all session data.
         */
        public function destroy_session()
        {
            $this->delete_session($this->_customer_id);
            $this->forget_session();
        }

        /**
         * Forget all session data without destroying it.
         */
        public function forget_session()
        {
            tbk_setcookie($this->_cookie, '', time() - YEAR_IN_SECONDS, $this->use_secure_cookie(), TRUE);


            // TODO: REMOVE ITEMS?

            $this->_data        = [];
            $this->_dirty       = FALSE;
            $this->_customer_id = $this->generate_customer_id();
        }

        /**
         * When a user is logged out, ensure they have a unique nonce to manage cart and more using the customer/session ID.
         * This filter runs everything `wp_verify_nonce()` and `wp_create_nonce()` gets called.
         *
         * @param int    $uid    User ID.
         * @param string $action The nonce action.
         *
         * @return int|string
         */
        public function maybe_update_nonce_user_logged_out($uid, $action)
        {
            if (strpos($action, 'tbk') === 0) {
                return $this->has_session() && $this->_customer_id ? $this->_customer_id : $uid;
            }

            return $uid;
        }

        /**
         * Cleanup session data from the database and clear caches.
         */
        public function cleanup_sessions()
        {
            global $wpdb;

            $wpdb->query($wpdb->prepare("DELETE FROM $this->_table WHERE session_expiry < %d", time()));
        }

        /**
         * Returns the session.
         *
         * @param string $customer_id Customer ID.
         * @param mixed  $default     Default session value.
         *
         * @return string|array
         */
        public function get_session($customer_id, $default = FALSE)
        {
            global $wpdb;

            if (defined('WP_SETUP_CONFIG')) {
                return FALSE;
            }

            $value = $wpdb->get_var($wpdb->prepare("SELECT session_value FROM $this->_table WHERE session_key = %s", $customer_id));

            if (is_null($value)) {
                $value = $default;
            }

            return maybe_unserialize($value);
        }


        /**
         * Delete the session from the cache and database.
         *
         * @param int $customer_id Customer ID.
         */
        public function delete_session($customer_id)
        {
            global $wpdb;

            $wpdb->delete(
                $this->_table,
                [
                    'session_key' => $customer_id,
                ]
            );
        }

        /**
         * Update the session expiry timestamp.
         *
         * @param string $customer_id Customer ID.
         * @param int    $timestamp   Timestamp to expire the cookie.
         */
        public function update_session_timestamp($customer_id, $timestamp)
        {
            global $wpdb;

            $wpdb->update(
                $this->_table,
                [
                    'session_expiry' => $timestamp,
                ],
                [
                    'session_key' => $customer_id,
                ],
                [
                    '%d',
                ]
            );
        }

    }

    /**
     * Set a cookie - wrapper for setcookie using WP constants.
     *
     * @param string  $name     Name of the cookie being set.
     * @param string  $value    Value of the cookie.
     * @param integer $expire   Expiry of the cookie.
     * @param bool    $secure   Whether the cookie should be served only over https.
     * @param bool    $httponly Whether the cookie is only accessible over HTTP, not scripting languages like JavaScript.
     */
    function tbk_setcookie($name, $value, $expire = 0, $secure = FALSE, $httponly = FALSE)
    {
        if (!headers_sent()) {
            /**
             * Controls the options to be specified when setting the cookie.
             *
             * @see   https://www.php.net/manual/en/function.setcookie.php
             *
             * @param array  $cookie_options Cookie options.
             * @param string $name           Cookie name.
             * @param string $value          Cookie value.
             */
            $options = apply_filters(
                'tbk_set_cookie_options',
                [
                    'expires'  => $expire,
                    'secure'   => $secure,
                    'path'     => COOKIEPATH ? COOKIEPATH : '/',
                    'domain'   => COOKIE_DOMAIN,
                    /**
                     * Controls whether the cookie should only be accessible via the HTTP protocol, or if it should also be
                     * accessible to Javascript.
                     *
                     * @see   https://www.php.net/manual/en/function.setcookie.php
                     *
                     * @param bool   $httponly If the cookie should only be accessible via the HTTP protocol.
                     * @param string $name     Cookie name.
                     * @param string $value    Cookie value.
                     * @param int    $expire   When the cookie should expire.
                     * @param bool   $secure   If the cookie should only be served over HTTPS.
                     */
                    'httponly' => apply_filters('tbk_cookie_httponly', $httponly, $name, $value, $expire, $secure),
                ],
                $name,
                $value
            );

            if (PHP_VERSION_ID >= 70300) {
                setcookie($name, $value, $options);
            } else {
                setcookie($name, $value, $options['expires'], $options['path'], $options['domain'], $options['secure'], $options['httponly']);
            }
        } elseif (defined('WP_DEBUG') && WP_DEBUG) {
            headers_sent($file, $line);
            trigger_error("{$name} cookie cannot be set - headers already sent by {$file} on line {$line}", E_USER_NOTICE);
        }
    }
}