<?php

namespace VSHM;

use Spatie\Period\Boundaries;
use Spatie\Period\Period;
use Spatie\Period\PeriodCollection;
use Spatie\Period\Precision;

defined('ABSPATH') || exit;

if (!class_exists(Tools::class)) {

    /**
     * Class Tools
     *
     * @package VSHM
     * @author  VonStroheim
     */
    class Tools
    {
        /**
         * @param PeriodCollection        $mainPeriodCollection
         * @param PeriodCollection|Period $periods
         *
         * @return PeriodCollection
         */
        public static function periodSubtract(PeriodCollection $mainPeriodCollection, $periods): PeriodCollection
        {
            if ($periods instanceof Period) {
                $periods = PeriodCollection::make($periods);
            }

            if ($periods->isEmpty()) {
                return $mainPeriodCollection;
            }

            $collection = [];

            foreach ($mainPeriodCollection as $period) {

                if ($period->getIncludedStart() == $period->getIncludedEnd()) {
                    continue;
                }

                $overlapping_periods = [];
                foreach ($periods as $inner_period) {
                    if ($period->overlapsWith($inner_period)) {
                        $overlapping_periods[] = $inner_period;
                    }
                }

                if (!empty($overlapping_periods)) {

                    $boundPeriod = Period::make(
                        $period->getStart(),
                        $period->getEnd(),
                        Precision::SECOND,
                        Boundaries::EXCLUDE_NONE
                    );

                    foreach ($boundPeriod->diff(...$overlapping_periods) as $item) {

                        if ($item->getIncludedStart() == $item->getIncludedEnd()) {
                            continue;
                        }

                        $collection[] = Period::make(
                            $item->getStart(),
                            $item->getEnd(),
                            Precision::SECOND,
                            Boundaries::EXCLUDE_ALL
                        );
                    }

                } else {
                    $collection[] = $period;
                }

            }


            return PeriodCollection::make(...$collection);
        }

        /**
         * Insert a value or key/value pair after a specific key in an array.  If key doesn't exist, value is appended
         * to the end of the array.
         *
         * @credits https://gist.github.com/wpscholar/0deadce1bbfa4adb4e4c
         *
         * @param array  $array
         * @param string $key
         * @param array  $new
         *
         * @return array
         */
        public static function array_insert_after(array $array, $key, array $new): array
        {
            $keys  = array_keys($array);
            $index = array_search($key, $keys, TRUE);
            $pos   = FALSE === $index ? count($array) : $index + 1;

            return array_merge(array_slice($array, 0, $pos), $new, array_slice($array, $pos));
        }

        /**
         * @param array $file
         *
         * @return string
         */
        public static function file_hash($file): string
        {
            return md5(md5_file($file['file']) . $file['file']);
        }

        public static function starts_with($string, $startString): bool
        {
            return (strpos($string, $startString) === 0);
        }

        /**
         * @param $string
         *
         * @return string
         */
        public static function mb_strtoupper($string): string
        {
            if (function_exists('mb_strtoupper')) {
                return mb_strtoupper($string, 'UTF-8');
            }

            return strtoupper($string);
        }

        /**
         * @param $string
         *
         * @return string
         */
        public static function mb_strtolower($string): string
        {
            if (function_exists('mb_strtolower')) {
                return mb_strtolower($string, 'UTF-8');
            }

            return strtolower($string);
        }

        /**
         * @param $string
         * @param $to
         * @param $charset
         *
         * @return string
         */
        public static function mb_convert_encoding($string, $to, $charset)
        {
            if (function_exists('mb_convert_encoding')) {
                return mb_convert_encoding($string, $to, $charset);
            }
            if (defined('WP_DEBUG') && WP_DEBUG) {
                trigger_error('Multibyte Extensions not installed, this may cause issues -- thrown when processing PayPal IPN --', E_USER_NOTICE);
            }

            return $string;
        }

        /**
         * @param array $array1
         * @param array $array2
         *
         * @return array
         */
        public static function array_merge_recursive_distinct(array &$array1, array &$array2): array
        {
            $merged = $array1;
            foreach ($array2 as $key => &$value) {
                if (is_array($value) && isset ($merged [ $key ]) && is_array($merged [ $key ])) {
                    $merged [ $key ] = static::array_merge_recursive_distinct($merged [ $key ], $value);
                } else {
                    $merged [ $key ] = $value;
                }
            }

            return $merged;
        }

        public static function subscribe_classes_in_dir($path, $classRoot, $subscribe_fn = 'bootstrap', $exclude = []): void
        {
            $classes = glob($path . '*.php');
            foreach ($classes as $class) {
                $classname = $classRoot . basename($class, '.php');
                if (is_callable([$classname, $subscribe_fn])
                    && !in_array(basename($class), $exclude, TRUE)) {
                    $classname::$subscribe_fn();
                }
            }
        }

        public static function subscribe_classes_in_dir_recursive($path, $classRoot, $subscribe_fn = 'bootstrap', $exclude = []): void
        {
            $classes = new \RecursiveDirectoryIterator($path, \FilesystemIterator::SKIP_DOTS);
            $classes = new \RecursiveIteratorIterator($classes);
            $classes = new \RegexIterator($classes, '/\.php$/i');
            foreach ($classes as $class) {

                $base      = str_replace(__DIR__ . DIRECTORY_SEPARATOR . 'Settings' . DIRECTORY_SEPARATOR, '', $class->getPathname());
                $base      = str_replace('/', '\\', $base);
                $classname = $classRoot . str_replace('.php', '', $base);

                if (is_callable([$classname, $subscribe_fn])
                    && !in_array(basename($class), $exclude, TRUE)
                ) {
                    $classname::$subscribe_fn();
                }
            }
        }

        /**
         * Maps MIME types to relative file extensions.
         *
         * @credits https://gist.github.com/alexcorvi/df8faecb59e86bee93411f6a7967df2c
         *
         * @param string $mime
         *
         * @return bool|mixed
         */
        public static function mime_to_ext($mime)
        {
            $mime_map = [
                'video/3gpp2'                                                               => '3g2',
                'video/3gp'                                                                 => '3gp',
                'video/3gpp'                                                                => '3gp',
                'application/x-compressed'                                                  => '7zip',
                'audio/x-acc'                                                               => 'aac',
                'audio/ac3'                                                                 => 'ac3',
                'application/postscript'                                                    => 'ai',
                'audio/x-aiff'                                                              => 'aif',
                'audio/aiff'                                                                => 'aif',
                'audio/x-au'                                                                => 'au',
                'video/x-msvideo'                                                           => 'avi',
                'video/msvideo'                                                             => 'avi',
                'video/avi'                                                                 => 'avi',
                'application/x-troff-msvideo'                                               => 'avi',
                'application/macbinary'                                                     => 'bin',
                'application/mac-binary'                                                    => 'bin',
                'application/x-binary'                                                      => 'bin',
                'application/x-macbinary'                                                   => 'bin',
                'image/bmp'                                                                 => 'bmp',
                'image/x-bmp'                                                               => 'bmp',
                'image/x-bitmap'                                                            => 'bmp',
                'image/x-xbitmap'                                                           => 'bmp',
                'image/x-win-bitmap'                                                        => 'bmp',
                'image/x-windows-bmp'                                                       => 'bmp',
                'image/ms-bmp'                                                              => 'bmp',
                'image/x-ms-bmp'                                                            => 'bmp',
                'application/bmp'                                                           => 'bmp',
                'application/x-bmp'                                                         => 'bmp',
                'application/x-win-bitmap'                                                  => 'bmp',
                'application/cdr'                                                           => 'cdr',
                'application/coreldraw'                                                     => 'cdr',
                'application/x-cdr'                                                         => 'cdr',
                'application/x-coreldraw'                                                   => 'cdr',
                'image/cdr'                                                                 => 'cdr',
                'image/x-cdr'                                                               => 'cdr',
                'zz-application/zz-winassoc-cdr'                                            => 'cdr',
                'application/mac-compactpro'                                                => 'cpt',
                'application/pkix-crl'                                                      => 'crl',
                'application/pkcs-crl'                                                      => 'crl',
                'application/x-x509-ca-cert'                                                => 'crt',
                'application/pkix-cert'                                                     => 'crt',
                'text/css'                                                                  => 'css',
                'text/x-comma-separated-values'                                             => 'csv',
                'text/comma-separated-values'                                               => 'csv',
                'application/vnd.msexcel'                                                   => 'csv',
                'application/x-director'                                                    => 'dcr',
                'application/vnd.openxmlformats-officedocument.wordprocessingml.document'   => 'docx',
                'application/x-dvi'                                                         => 'dvi',
                'message/rfc822'                                                            => 'eml',
                'application/x-msdownload'                                                  => 'exe',
                'video/x-f4v'                                                               => 'f4v',
                'audio/x-flac'                                                              => 'flac',
                'video/x-flv'                                                               => 'flv',
                'image/gif'                                                                 => 'gif',
                'application/gpg-keys'                                                      => 'gpg',
                'application/x-gtar'                                                        => 'gtar',
                'application/x-gzip'                                                        => 'gzip',
                'application/mac-binhex40'                                                  => 'hqx',
                'application/mac-binhex'                                                    => 'hqx',
                'application/x-binhex40'                                                    => 'hqx',
                'application/x-mac-binhex40'                                                => 'hqx',
                'text/html'                                                                 => 'html',
                'image/x-icon'                                                              => 'ico',
                'image/x-ico'                                                               => 'ico',
                'image/vnd.microsoft.icon'                                                  => 'ico',
                'text/calendar'                                                             => 'ics',
                'application/java-archive'                                                  => 'jar',
                'application/x-java-application'                                            => 'jar',
                'application/x-jar'                                                         => 'jar',
                'image/jp2'                                                                 => 'jp2',
                'video/mj2'                                                                 => 'jp2',
                'image/jpx'                                                                 => 'jp2',
                'image/jpm'                                                                 => 'jp2',
                'image/jpeg'                                                                => 'jpeg',
                'image/pjpeg'                                                               => 'jpeg',
                'application/x-javascript'                                                  => 'js',
                'application/json'                                                          => 'json',
                'text/json'                                                                 => 'json',
                'application/vnd.google-earth.kml+xml'                                      => 'kml',
                'application/vnd.google-earth.kmz'                                          => 'kmz',
                'text/x-log'                                                                => 'log',
                'audio/x-m4a'                                                               => 'm4a',
                'application/vnd.mpegurl'                                                   => 'm4u',
                'audio/midi'                                                                => 'mid',
                'application/vnd.mif'                                                       => 'mif',
                'video/quicktime'                                                           => 'mov',
                'video/x-sgi-movie'                                                         => 'movie',
                'audio/mpeg'                                                                => 'mp3',
                'audio/mpg'                                                                 => 'mp3',
                'audio/mpeg3'                                                               => 'mp3',
                'audio/mp3'                                                                 => 'mp3',
                'video/mp4'                                                                 => 'mp4',
                'video/mpeg'                                                                => 'mpeg',
                'application/oda'                                                           => 'oda',
                'audio/ogg'                                                                 => 'ogg',
                'video/ogg'                                                                 => 'ogg',
                'application/ogg'                                                           => 'ogg',
                'application/x-pkcs10'                                                      => 'p10',
                'application/pkcs10'                                                        => 'p10',
                'application/x-pkcs12'                                                      => 'p12',
                'application/x-pkcs7-signature'                                             => 'p7a',
                'application/pkcs7-mime'                                                    => 'p7c',
                'application/x-pkcs7-mime'                                                  => 'p7c',
                'application/x-pkcs7-certreqresp'                                           => 'p7r',
                'application/pkcs7-signature'                                               => 'p7s',
                'application/pdf'                                                           => 'pdf',
                'application/octet-stream'                                                  => 'pdf',
                'application/x-x509-user-cert'                                              => 'pem',
                'application/x-pem-file'                                                    => 'pem',
                'application/pgp'                                                           => 'pgp',
                'application/x-httpd-php'                                                   => 'php',
                'application/php'                                                           => 'php',
                'application/x-php'                                                         => 'php',
                'text/php'                                                                  => 'php',
                'text/x-php'                                                                => 'php',
                'application/x-httpd-php-source'                                            => 'php',
                'image/png'                                                                 => 'png',
                'image/x-png'                                                               => 'png',
                'application/powerpoint'                                                    => 'ppt',
                'application/vnd.ms-powerpoint'                                             => 'ppt',
                'application/vnd.ms-office'                                                 => 'ppt',
                'application/msword'                                                        => 'ppt',
                'application/vnd.openxmlformats-officedocument.presentationml.presentation' => 'pptx',
                'application/x-photoshop'                                                   => 'psd',
                'image/vnd.adobe.photoshop'                                                 => 'psd',
                'audio/x-realaudio'                                                         => 'ra',
                'audio/x-pn-realaudio'                                                      => 'ram',
                'application/x-rar'                                                         => 'rar',
                'application/rar'                                                           => 'rar',
                'application/x-rar-compressed'                                              => 'rar',
                'audio/x-pn-realaudio-plugin'                                               => 'rpm',
                'application/x-pkcs7'                                                       => 'rsa',
                'text/rtf'                                                                  => 'rtf',
                'text/richtext'                                                             => 'rtx',
                'video/vnd.rn-realvideo'                                                    => 'rv',
                'application/x-stuffit'                                                     => 'sit',
                'application/smil'                                                          => 'smil',
                'text/srt'                                                                  => 'srt',
                'image/svg+xml'                                                             => 'svg',
                'application/x-shockwave-flash'                                             => 'swf',
                'application/x-tar'                                                         => 'tar',
                'application/x-gzip-compressed'                                             => 'tgz',
                'image/tiff'                                                                => 'tiff',
                'text/plain'                                                                => 'txt',
                'text/x-vcard'                                                              => 'vcf',
                'application/videolan'                                                      => 'vlc',
                'text/vtt'                                                                  => 'vtt',
                'audio/x-wav'                                                               => 'wav',
                'audio/wave'                                                                => 'wav',
                'audio/wav'                                                                 => 'wav',
                'application/wbxml'                                                         => 'wbxml',
                'video/webm'                                                                => 'webm',
                'audio/x-ms-wma'                                                            => 'wma',
                'application/wmlc'                                                          => 'wmlc',
                'video/x-ms-wmv'                                                            => 'wmv',
                'video/x-ms-asf'                                                            => 'wmv',
                'application/xhtml+xml'                                                     => 'xhtml',
                'application/excel'                                                         => 'xl',
                'application/msexcel'                                                       => 'xls',
                'application/x-msexcel'                                                     => 'xls',
                'application/x-ms-excel'                                                    => 'xls',
                'application/x-excel'                                                       => 'xls',
                'application/x-dos_ms_excel'                                                => 'xls',
                'application/xls'                                                           => 'xls',
                'application/x-xls'                                                         => 'xls',
                'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'         => 'xlsx',
                'application/vnd.ms-excel'                                                  => 'xlsx',
                'application/xml'                                                           => 'xml',
                'text/xml'                                                                  => 'xml',
                'text/xsl'                                                                  => 'xsl',
                'application/xspf+xml'                                                      => 'xspf',
                'application/x-compress'                                                    => 'z',
                'application/x-zip'                                                         => 'zip',
                'application/zip'                                                           => 'zip',
                'application/x-zip-compressed'                                              => 'zip',
                'application/s-compressed'                                                  => 'zip',
                'multipart/x-zip'                                                           => 'zip',
                'text/x-scriptzsh'                                                          => 'zsh',
            ];

            return isset($mime_map[ $mime ]) === TRUE ? $mime_map[ $mime ] : FALSE;
        }

        /**
         * Makes an array ready for HTML attribute insertion.
         *
         * @param array $params
         *
         * @return string
         */
        public static function prepare_attribute_params(array $params)
        {
            return htmlspecialchars(json_encode($params), ENT_QUOTES);
        }

        /**
         * Checks if an array is an associative array.
         *
         * @param array $arr
         *
         * @return bool
         */
        public static function array_is_assoc(array $arr)
        {
            if ([] === $arr) {
                return FALSE;
            }

            return array_keys($arr) !== range(0, count($arr) - 1);
        }

        /**
         * @param $string
         *
         * @return array|string|string[]
         */
        public static function escapeJavaScriptText($string)
        {
            return str_replace(['"', "\n"], ['\"', '\n'], addcslashes(str_replace("\r", '', (string)$string), "\0..\37'\\"));
        }

        /**
         * @param $class
         *
         * @return false|string
         */
        public static function get_short_classname($class)
        {
            return substr(strrchr('\\' . $class, '\\'), 1);
        }

        /**
         * Generates a secure token
         * (source https://gist.github.com/raveren/5555297)
         *
         * @param string $type
         * @param int    $length
         * @param string $prefix
         *
         * @return string
         */
        public static function generate_token($type = 'alnum', $length = 32, $prefix = '')
        {
            switch ($type) {
                case 'alnum':
                    $pool = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
                    break;
                case 'alpha':
                    $pool = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
                    break;
                case 'hexdec':
                    $pool = '0123456789abcdef';
                    break;
                case 'numeric':
                    $pool = '0123456789';
                    break;
                case 'nozero':
                    $pool = '123456789';
                    break;
                case 'distinct':
                    $pool = '2345679ACDEFHJKLMNPRSTUVWXYZ';
                    break;
                case 'alnum_upper':
                    $pool = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';
                    break;
                default:
                    $pool = (string)$type;
                    break;
            }

            $crypto_rand_secure = static function ($min, $max) {
                $range = $max - $min;
                if ($range < 0) {
                    return $min;
                }
                $log    = log($range, 2);
                $bytes  = (int)($log / 8) + 1; // length in bytes
                $bits   = (int)$log + 1; // length in bits
                $filter = (1 << $bits) - 1; // set all lower bits to 1
                do {
                    $rnd = hexdec(bin2hex(random_bytes($bytes)));
                    $rnd &= $filter; // discard irrelevant bits
                } while ($rnd >= $range);

                return $min + $rnd;
            };

            $token = '';
            $max   = strlen($pool);
            for ($i = 0; $i < $length; $i++) {
                $token .= $pool[ $crypto_rand_secure(0, $max) ];
            }

            return $prefix . $token;
        }

        /**
         * Serialize data, if needed.
         *
         * Unlike WordPress core function, this one
         * prevents double-serialization.
         *
         * @param $data
         *
         * @return string
         */
        public static function maybe_serialize($data)
        {
            if (is_array($data) || is_object($data)) {
                return serialize($data);
            }

            return $data;
        }

        /**
         * Changes a key in the array while keeping its order.
         *
         * @param array  $array
         * @param string $old_k
         * @param string $new_k
         *
         * @return array
         */
        public static function change_array_key($array, $old_k, $new_k)
        {
            if (!array_key_exists($old_k, $array)) {
                return $array;
            }
            $keys = array_keys($array);

            $keys[ array_search($old_k, $keys, TRUE) ] = $new_k;

            return array_combine($keys, $array);
        }

        /**
         * Looks for a JSON string inside a generic string
         * which is supposed to contain one JSON string only.
         *
         * Returns any decoded finding.
         *
         * @param $string
         *
         * @return array|bool|mixed|object
         */
        public static function looking_for_json($string)
        {
            $regex   = "/(\{.*?\})/s";
            $matches = array();
            preg_match($regex, $string, $matches);

            return isset($matches[0]) ? self::json_validate($matches[0]) : FALSE;
        }

        /**
         * Validates a JSON string
         *
         * @param $string
         *
         * @return array|bool|mixed|object
         */
        public static function json_validate($string)
        {
            $result = json_decode($string, FALSE);

            switch (json_last_error()) {
                case JSON_ERROR_NONE:
                    $error = '';
                    break;
                case JSON_ERROR_DEPTH:
                    $error = 'The maximum stack depth has been exceeded.';
                    break;
                case JSON_ERROR_STATE_MISMATCH:
                    $error = 'Invalid or malformed JSON.';
                    break;
                case JSON_ERROR_CTRL_CHAR:
                    $error = 'Control character error, possibly incorrectly encoded.';
                    break;
                case JSON_ERROR_SYNTAX:
                    $error = 'Syntax error, malformed JSON.';
                    break;
                // PHP >= 5.3.3
                case JSON_ERROR_UTF8:
                    $error = 'Malformed UTF-8 characters, possibly incorrectly encoded.';
                    break;
                // PHP >= 5.5.0
                case JSON_ERROR_RECURSION:
                    $error = 'One or more recursive references in the value to be encoded.';
                    break;
                // PHP >= 5.5.0
                case JSON_ERROR_INF_OR_NAN:
                    $error = 'One or more NAN or INF values in the value to be encoded.';
                    break;
                case JSON_ERROR_UNSUPPORTED_TYPE:
                    $error = 'A value of a type that cannot be encoded was given.';
                    break;
                default:
                    $error = 'Unknown JSON error occurred.';
                    break;
            }

            if (!empty($error)) {
                $result = FALSE;
                if (defined('WP_DEBUG') && WP_DEBUG) {
                    $error .= ' Original string: ' . $string;
                    trigger_error(esc_html($error));
                }
            }

            return $result;
        }

        /**
         * Useful debug function to stringify a variable dump.
         *
         * @param $var
         *
         * @return string
         */
        public static function stringify_dump($var): string
        {
            ob_start();
            var_dump($var);

            return ob_get_clean();
        }

        /**
         * Useful debug function to log a variable dump.
         *
         * @param $var
         */
        public static function log_dump($var): void
        {
            error_log(self::stringify_dump($var));
        }

        /**
         * Checks what kind of request the script is executing in WP context.
         *
         * @param string $type
         *
         * @return bool
         */
        public static function is_request(string $type): bool
        {
            switch ($type) {
                case 'admin':
                    return is_admin();
                case 'ajax':
                    return defined('DOING_AJAX') && DOING_AJAX;
                case 'rest':
                    if ((defined('REST_REQUEST') && REST_REQUEST)
                        || (isset($_GET['rest_route']) && strpos(sanitize_text_field($_GET['rest_route']), REST_Controller::NAME_SPACE))) {
                        return TRUE;
                    }
                    $is_rest = FALSE;
                    if (!empty($_SERVER['REQUEST_URI'])) {
                        $rest_url     = self::get_rest_url(get_current_blog_id(), '/');
                        $rest_path    = trim(parse_url($rest_url, PHP_URL_PATH), '/');
                        $request_path = trim(sanitize_text_field($_SERVER['REQUEST_URI']), '/');
                        $is_rest      = (strpos($request_path, $rest_path) === 0);
                    }

                    return $is_rest;
                case 'cron':
                    return defined('DOING_CRON') && DOING_CRON;
                case 'frontend':
                    return (!is_admin() || (defined('DOING_AJAX') && DOING_AJAX))
                        && (!defined('DOING_CRON') || !DOING_CRON)
                        && (!defined('REST_REQUEST') || !REST_REQUEST);
            }

            return FALSE;
        }

        /**
         * This function is a bypass of WordPress original function
         * as it calls WP_Rewrite and we need it earlier.
         *
         * Todo: check when WordPress will drop the WP_Rewrite call, as it is not actually necessary.
         *
         * @param null   $blog_id
         * @param string $path
         * @param string $scheme
         *
         * @return mixed|void
         */
        public static function get_rest_url($blog_id = NULL, $path = '/', $scheme = 'rest')
        {
            if (empty($path)) {
                $path = '/';
            }

            $path = '/' . ltrim($path, '/');

            if (get_option('permalink_structure') || (is_multisite() && get_blog_option($blog_id, 'permalink_structure'))) {

                if (preg_match('#^/*index.php#', get_option('permalink_structure'))) {
                    $url = get_home_url($blog_id, 'index.php/' . rest_get_url_prefix(), $scheme);
                } else {
                    $url = get_home_url($blog_id, rest_get_url_prefix(), $scheme);
                }

                $url .= $path;
            } else {
                $url = trailingslashit(get_home_url($blog_id, '', $scheme));
                if ('index.php' !== substr($url, 9)) {
                    $url .= 'index.php';
                }
                $url = add_query_arg('rest_route', $path, $url);
            }

            if (isset($_SERVER['SERVER_NAME']) && is_ssl() && $_SERVER['SERVER_NAME'] === parse_url(get_home_url($blog_id), PHP_URL_HOST)) {
                $url = set_url_scheme($url, 'https');
            }

            if (is_admin() && force_ssl_admin()) {
                $url = set_url_scheme($url, 'https');
            }

            return apply_filters('rest_url', $url, $path, $blog_id, $scheme);
        }

        /**
         * Helper to enqueue styles.
         *
         * @param string $handle
         * @param string $path
         * @param array  $deps
         */
        public static function enqueue_style(string $handle, string $path, array $deps = []): void
        {
            wp_enqueue_style($handle,
                vshm()->plugin['URL'] . $path,
                $deps,
                filemtime(vshm()->plugin['DIR'] . implode(DIRECTORY_SEPARATOR, explode('/', $path)))
            );
        }

        /**
         * Helper to enqueue scripts.
         *
         * @param string $handle
         * @param string $path
         * @param array  $deps
         * @param bool   $footer
         */
        public static function enqueue_script(string $handle, string $path, array $deps = [], bool $footer = FALSE): void
        {
            if ($path[0] . $path[1] !== '//') {
                wp_enqueue_script($handle,
                    vshm()->plugin['URL'] . $path,
                    $deps,
                    filemtime(vshm()->plugin['DIR'] . implode(DIRECTORY_SEPARATOR, explode('/', $path))),
                    $footer
                );
            } else {
                wp_enqueue_script($handle,
                    $path,
                    $deps,
                    NULL,
                    $footer
                );
            }
        }

        /**
         * Returns internationalized weekdays labels.
         *
         * @param string $format
         *
         * @return array
         */
        public static function i18n_weekdays_labels(string $format = 'l'): array
        {
            $return = [];
            for ($i = 0; $i < 7; $i++) {
                $return[ $i ] = wp_date($format, strtotime("Sunday +{$i} days"), new \DateTimeZone('UTC'));
            }

            return $return;
        }

        /**
         * Returns internationalized months labels.
         *
         * @param string $format
         *
         * @return array
         */
        public static function i18n_months_labels($format = 'F')
        {
            $return = [];
            for ($i = 0; $i < 12; $i++) {
                $return[ $i ] = wp_date($format, strtotime("1st January +{$i} months"), new \DateTimeZone('UTC'));
            }

            return $return;
        }

        /**
         * @return string
         */
        public static function get_ip_address()
        {
            foreach (['HTTP_CLIENT_IP', 'HTTP_X_FORWARDED_FOR', 'HTTP_X_FORWARDED', 'HTTP_X_CLUSTER_CLIENT_IP', 'HTTP_FORWARDED_FOR', 'HTTP_FORWARDED', 'REMOTE_ADDR'] as $key) {
                if (array_key_exists($key, $_SERVER) === TRUE) {
                    foreach (explode(',', $_SERVER[ $key ]) as $ip) {
                        $ip = trim($ip);
                        if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE) !== FALSE) {
                            return $ip;
                        }
                    }
                }
            }

            return NULL;
        }

        /**
         * Checks whether a dark theme may be appropriate.
         *
         * @param $hexColor
         *
         * @return bool
         */
        public static function requiresDarkTheme($hexColor)
        {
            $R1           = hexdec(substr($hexColor, 1, 2));
            $G1           = hexdec(substr($hexColor, 3, 2));
            $B1           = hexdec(substr($hexColor, 5, 2));
            $blackColor   = "#000000";
            $R2BlackColor = hexdec(substr($blackColor, 1, 2));
            $G2BlackColor = hexdec(substr($blackColor, 3, 2));
            $B2BlackColor = hexdec(substr($blackColor, 5, 2));

            $L1 = 0.2126 * (($R1 / 255) ** 2.2) +
                0.7152 * (($G1 / 255) ** 2.2) +
                0.0722 * (($B1 / 255) ** 2.2);

            $L2 = 0.2126 * (($R2BlackColor / 255) ** 2.2) +
                0.7152 * (($G2BlackColor / 255) ** 2.2) +
                0.0722 * (($B2BlackColor / 255) ** 2.2);

            if ($L1 > $L2) {
                $contrastRatio = (int)(($L1 + 0.05) / ($L2 + 0.05));
            } else {
                $contrastRatio = (int)(($L2 + 0.05) / ($L1 + 0.05));
            }

            return !($contrastRatio > 5);
        }

        /**
         * Increases or decreases the brightness of a color by a percentage of the current brightness.
         *
         * @param string $hexCode       Supported formats: `#FFF`, `#FFFFFF`, `FFF`, `FFFFFF`
         * @param float  $adjustPercent A number between -1 and 1. E.g. 0.3 = 30% lighter; -0.4 = 40% darker.
         *
         * @return  string
         */
        public static function adjustBrightness($hexCode, $adjustPercent)
        {
            $hexCode = ltrim($hexCode, '#');

            if (strlen($hexCode) === 3) {
                $hexCode = $hexCode[0] . $hexCode[0] . $hexCode[1] . $hexCode[1] . $hexCode[2] . $hexCode[2];
            }

            $hexCode = array_map('hexdec', str_split($hexCode, 2));

            foreach ($hexCode as & $color) {
                $adjustableLimit = $adjustPercent < 0 ? $color : 255 - $color;
                $adjustAmount    = ceil($adjustableLimit * $adjustPercent);

                $color = str_pad(dechex($color + $adjustAmount), 2, '0', STR_PAD_LEFT);
            }

            return '#' . implode($hexCode);
        }

        /**
         * Returns a list of WordPress timezones.
         *
         * @param string $locale
         *
         * @return string[]
         */
        public static function timezone_list($locale = NULL)
        {
            static $mo_loaded = FALSE, $locale_loaded = NULL;

            if (!$mo_loaded || $locale !== $locale_loaded) {
                if ($locale) {
                    $locale_loaded = $locale;
                } else {
                    $locale_loaded = get_locale();
                }
                $mofile = WP_LANG_DIR . '/continents-cities-' . $locale_loaded . '.mo';
                unload_textdomain('continents-cities');
                load_textdomain('continents-cities', $mofile);
                $mo_loaded = TRUE;
            }

            $zonen = [];

            foreach (timezone_identifiers_list() as $zone) {
                $zone = explode('/', $zone);

                $exists    = [
                    0 => isset($zone[0]) && $zone[0],
                    1 => isset($zone[1]) && $zone[1],
                    2 => isset($zone[2]) && $zone[2],
                ];
                $exists[3] = ($exists[0] && 'Etc' !== $zone[0]);
                $exists[4] = ($exists[1] && $exists[3]);
                $exists[5] = ($exists[2] && $exists[3]);

                $zonen[] = [
                    'continent'   => $exists[0] ? $zone[0] : '',
                    'city'        => $exists[1] ? $zone[1] : '',
                    'subcity'     => $exists[2] ? $zone[2] : '',
                    't_continent' => $exists[3] ? translate(str_replace('_', ' ', $zone[0]), 'continents-cities') : '',
                    't_city'      => $exists[4] ? translate(str_replace('_', ' ', $zone[1]), 'continents-cities') : '',
                    't_subcity'   => $exists[5] ? translate(str_replace('_', ' ', $zone[2]), 'continents-cities') : ''
                ];

                usort($zonen, '_wp_timezone_choice_usort_callback');


            }

            return $zonen;
        }

        /**
         * A list of almost all currencies
         *
         * @return array
         */
        public static function getCurrencies($code = NULL)
        {
            $currencies = [
                'AED' => [
                    'label'   => 'United Arab Emirates Dirham',
                    'format'  => 'after',
                    'locale'  => 'ar_AE',
                    'symbol'  => '&#x62f;&#x2e;&#x625;',
                    'decimal' => TRUE
                ],
                'AFN' => [
                    'label'   => 'Afghan Afghani',
                    'format'  => 'after',
                    'locale'  => 'fa_AF',
                    'symbol'  => '&#1547;',
                    'decimal' => TRUE
                ],
                'ALL' => [
                    'label'   => 'Albanian Lek',
                    'format'  => 'before',
                    'locale'  => 'sq_AL',
                    'symbol'  => 'Lek',
                    'decimal' => TRUE
                ],
                'AMD' => [
                    'label'   => 'Armenian Dram',
                    'format'  => 'before',
                    'locale'  => 'hy_AM',
                    'symbol'  => '&#1423;',
                    'decimal' => TRUE
                ],
                'ANG' => [
                    'label'   => 'Netherlands Antillean Gulden',
                    'format'  => 'before',
                    'locale'  => 'nl_SX',
                    'symbol'  => '&#402;',
                    'decimal' => TRUE
                ],
                'AOA' => [
                    'label'   => 'Angolan Kwanza',
                    'format'  => 'before',
                    'locale'  => 'pt_AO',
                    'symbol'  => 'Kz;',
                    'decimal' => TRUE
                ],
                'ARS' => [
                    'label'   => 'Argentine Peso',
                    'format'  => 'before',
                    'locale'  => 'es_AR',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'AUD' => [
                    'label'   => 'Australian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_AU',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'AWG' => [
                    'label'   => 'Aruban Florin',
                    'format'  => 'before',
                    'locale'  => 'nl_AW',
                    'symbol'  => '&#402;',
                    'decimal' => TRUE
                ],
                'AZN' => [
                    'label'   => 'Azerbaijani Manat',
                    'format'  => 'before',
                    'locale'  => 'az_Latn_AZ',
                    'symbol'  => '&#8380;',
                    'decimal' => TRUE
                ],
                'BAM' => [
                    'label'  => 'Bosnia & Herzegovina Convertible Mark',
                    'format' => 'before',
                    'locale' => 'hr_BA',
                    'symbol' => 'KM',
                ],
                'BBD' => [
                    'label'   => 'Barbadian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_BB',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'BDT' => [
                    'label'   => 'Bangladeshi Taka',
                    'format'  => 'before',
                    'locale'  => 'bn_BD',
                    'symbol'  => '&#2547;',
                    'decimal' => TRUE
                ],
                'BGN' => [
                    'label'   => 'Bulgarian Lev',
                    'format'  => 'after',
                    'locale'  => 'bg_BG',
                    'symbol'  => 'лв',
                    'decimal' => TRUE
                ],
                'BIF' => [
                    'label'   => 'Burundian Franc',
                    'format'  => 'before',
                    'locale'  => 'rn_BI',
                    'symbol'  => 'FBu',
                    'decimal' => FALSE
                ],
                'BMD' => [
                    'label'  => 'Bermudian Dollar',
                    'format' => 'before',
                    'locale' => 'en_BM',
                    'symbol' => '$',
                ],
                'BND' => [
                    'label'   => 'Brunei Dollar',
                    'format'  => 'before',
                    'locale'  => 'ms_Latn_BN',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'BOB' => [
                    'label'   => 'Bolivian Boliviano',
                    'format'  => 'before',
                    'locale'  => 'es_BO',
                    'symbol'  => 'Bs.',
                    'decimal' => TRUE
                ],
                'BRL' => [
                    'label'   => 'Brazilian Real',
                    'format'  => 'before',
                    'locale'  => 'pt_BR',
                    'symbol'  => 'R$',
                    'decimal' => TRUE
                ],
                'BSD' => [
                    'label'   => 'Bahamian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_BS',
                    'symbol'  => 'B$',
                    'decimal' => TRUE
                ],
                'BWP' => [
                    'label'   => 'Botswana Pula',
                    'format'  => 'before',
                    'locale'  => 'en_BW',
                    'symbol'  => 'P',
                    'decimal' => TRUE
                ],
                'BZD' => [
                    'label'   => 'Belize Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_BZ',
                    'symbol'  => 'BZ$',
                    'decimal' => TRUE
                ],
                'CAD' => [
                    'label'   => 'Canadian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_CA',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'CDF' => [
                    'label'   => 'Congolese Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_CD',
                    'symbol'  => 'FC',
                    'decimal' => TRUE
                ],
                'CHF' => [
                    'label'   => 'Swiss Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_CH',
                    'symbol'  => 'Fr',
                    'decimal' => TRUE
                ],
                'CLP' => [
                    'label'   => 'Chilean Peso',
                    'format'  => 'before',
                    'locale'  => 'es_CL',
                    'symbol'  => '$',
                    'decimal' => FALSE
                ],
                'CNY' => [
                    'label'   => 'Chinese Renminbi Yuan',
                    'format'  => 'before',
                    'locale'  => 'zh_Hans_CN',
                    'symbol'  => '&#165;',
                    'decimal' => TRUE
                ],
                'COP' => [
                    'label'   => 'Colombian Peso',
                    'format'  => 'before',
                    'locale'  => 'es_CO',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'CRC' => [
                    'label'   => 'Costa Rican Colón',
                    'format'  => 'before',
                    'locale'  => 'es_CR',
                    'symbol'  => '&#8353;',
                    'decimal' => TRUE
                ],
                'CVE' => [
                    'label'   => 'Cape Verdean Escudo',
                    'format'  => 'before',
                    'locale'  => 'pt_CV',
                    'symbol'  => 'Esc',
                    'decimal' => FALSE
                ],
                'CZK' => [
                    'label'   => 'Czech Koruna',
                    'format'  => 'after',
                    'locale'  => 'cs_CZ',
                    'symbol'  => 'Kč',
                    'decimal' => TRUE
                ],
                'DJF' => [
                    'label'   => 'Djiboutian Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_DJ',
                    'symbol'  => 'Fdj',
                    'decimal' => FALSE
                ],
                'DKK' => [
                    'label'   => 'Danish Krone',
                    'format'  => 'before',
                    'locale'  => 'da_DK',
                    'symbol'  => 'kr',
                    'decimal' => TRUE
                ],
                'DOP' => [
                    'label'   => 'Dominican Peso',
                    'format'  => 'before',
                    'locale'  => 'es_DO',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'DZD' => [
                    'label'   => 'Algerian Dinar',
                    'format'  => 'before',
                    'locale'  => 'fr_DZ',
                    'symbol'  => '&#1583;.&#1580;',
                    'decimal' => TRUE
                ],
                'EGP' => [
                    'label'   => 'Egyptian Pound',
                    'format'  => 'before',
                    'locale'  => 'ar_EG',
                    'symbol'  => 'E&pound;',
                    'decimal' => TRUE
                ],
                'ETB' => [
                    'label'   => 'Ethiopian Birr',
                    'format'  => 'before',
                    'locale'  => 'so_ET',
                    'symbol'  => 'Br',
                    'decimal' => TRUE
                ],
                'EUR' => [
                    'label'   => 'Euro',
                    'format'  => 'before',
                    'locale'  => '',
                    'symbol'  => '&euro;',
                    'decimal' => TRUE
                ],
                'FJD' => [
                    'label'   => 'Fijian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_FJ',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'FKP' => [
                    'label'   => 'Falkland Islands Pound',
                    'format'  => 'before',
                    'locale'  => 'en_FK',
                    'symbol'  => '&pound;',
                    'decimal' => TRUE
                ],
                'GBP' => [
                    'label'   => 'British Pound',
                    'format'  => 'before',
                    'locale'  => 'en_UK',
                    'symbol'  => '&pound;',
                    'decimal' => TRUE
                ],
                'GEL' => [
                    'label'   => 'Georgian Lari',
                    'format'  => 'before',
                    'locale'  => 'ka_GE',
                    'symbol'  => '&#4314;',
                    'decimal' => TRUE
                ],
                'GIP' => [
                    'label'   => 'Gibraltar Pound',
                    'format'  => 'before',
                    'locale'  => 'en_GI',
                    'symbol'  => '&pound;',
                    'decimal' => TRUE
                ],
                'GMD' => [
                    'label'   => 'Gambian Dalasi',
                    'format'  => 'before',
                    'locale'  => 'en_GM',
                    'symbol'  => 'D',
                    'decimal' => TRUE
                ],
                'GNF' => [
                    'label'   => 'Guinean Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_GN',
                    'symbol'  => 'FG',
                    'decimal' => FALSE
                ],
                'GTQ' => [
                    'label'   => 'Guatemalan Quetzal',
                    'format'  => 'before',
                    'locale'  => 'es_GT',
                    'symbol'  => 'Q',
                    'decimal' => TRUE
                ],
                'GYD' => [
                    'label'   => 'Guyanese Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_GY',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'HKD' => [
                    'label'   => 'Hong Kong Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_HK',
                    'symbol'  => 'HK$',
                    'decimal' => TRUE
                ],
                'HNL' => [
                    'label'   => 'Honduran Lempira',
                    'format'  => 'before',
                    'locale'  => 'es_HN',
                    'symbol'  => 'L',
                    'decimal' => TRUE
                ],
                'HRK' => [
                    'label'   => 'Croatian Kuna',
                    'format'  => 'before',
                    'locale'  => 'hr_HR',
                    'symbol'  => 'kn',
                    'decimal' => TRUE
                ],
                'HTG' => [
                    'label'   => 'Haitian Gourde',
                    'format'  => 'before',
                    'locale'  => 'fr_HT',
                    'symbol'  => 'G',
                    'decimal' => TRUE
                ],
                'HUF' => [
                    'label'   => 'Hungarian Forint',
                    'format'  => 'before',
                    'locale'  => 'hu_HU',
                    'symbol'  => 'Ft',
                    'decimal' => FALSE
                ],
                'IDR' => [
                    'label'   => 'Indonesian Rupiah',
                    'format'  => 'before',
                    'locale'  => 'id_ID',
                    'symbol'  => 'Rp',
                    'decimal' => TRUE
                ],
                'ILS' => [
                    'label'   => 'Israeli New Sheqel',
                    'format'  => 'before',
                    'locale'  => 'he_IL',
                    'symbol'  => '&#8362;',
                    'decimal' => TRUE
                ],
                'INR' => [
                    'label'   => 'Indian Rupee',
                    'format'  => 'before',
                    'locale'  => 'en_IN',
                    'symbol'  => '&#8377;',
                    'decimal' => TRUE
                ],
                'ISK' => [
                    'label'   => 'Icelandic Króna',
                    'format'  => 'before',
                    'locale'  => 'is_IS',
                    'symbol'  => 'kr',
                    'decimal' => FALSE
                ],
                'JMD' => [
                    'label'   => 'Jamaican Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_JM',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'JPY' => [
                    'label'   => 'Japanese Yen',
                    'format'  => 'before',
                    'locale'  => 'ja_JP',
                    'symbol'  => '&yen;',
                    'decimal' => FALSE
                ],
                'KES' => [
                    'label'   => 'Kenyan Shilling',
                    'format'  => 'before',
                    'locale'  => 'en_KE',
                    'symbol'  => 'KSh',
                    'decimal' => TRUE
                ],
                'KGS' => [
                    'label'   => 'Kyrgyzstani Som',
                    'format'  => 'before',
                    'locale'  => 'ru_KG',
                    'symbol'  => '&#1083;&#1074;',
                    'decimal' => TRUE
                ],
                'KHR' => [
                    'label'   => 'Cambodian Riel',
                    'format'  => 'before',
                    'locale'  => 'km_KH',
                    'symbol'  => '&#6107;',
                    'decimal' => TRUE
                ],
                'KMF' => [
                    'label'   => 'Comorian Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_KM',
                    'symbol'  => 'CF',
                    'decimal' => FALSE
                ],
                'KRW' => [
                    'label'   => 'South Korean Won',
                    'format'  => 'before',
                    'locale'  => 'ko_KR',
                    'symbol'  => '&#8361;',
                    'decimal' => FALSE
                ],
                'KYD' => [
                    'label'   => 'Cayman Islands Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_KY',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'KZT' => [
                    'label'   => 'Kazakhstani Tenge',
                    'format'  => 'before',
                    'locale'  => 'ru_KZ',
                    'symbol'  => '&#8376;',
                    'decimal' => TRUE
                ],
                'LAK' => [
                    'label'   => 'Lao Kipa',
                    'format'  => 'before',
                    'locale'  => 'lo_LA',
                    'symbol'  => '&#8365;',
                    'decimal' => TRUE
                ],
                'LBP' => [
                    'label'   => 'Lebanese Pound',
                    'format'  => 'after',
                    'locale'  => 'ar_LB',
                    'symbol'  => '&#1604;.&#1604;',
                    'decimal' => TRUE
                ],
                'LKR' => [
                    'label'   => 'Sri Lankan Rupee',
                    'format'  => 'before',
                    'locale'  => 'si_LK',
                    'symbol'  => '&#588;s',
                    'decimal' => TRUE
                ],
                'LRD' => [
                    'label'   => 'Liberian Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_LR',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'LSL' => [
                    'label'   => 'Lesotho Loti',
                    'format'  => 'before',
                    'locale'  => '',
                    'symbol'  => 'L',
                    'decimal' => TRUE
                ],
                'MAD' => [
                    'label'   => 'Moroccan Dirham',
                    'format'  => 'before',
                    'locale'  => 'ar_MA',
                    'symbol'  => '&#1583;.&#1605;.',
                    'decimal' => TRUE
                ],
                'MDL' => [
                    'label'   => 'Moldovan Leu',
                    'format'  => 'before',
                    'locale'  => 'ro_MD',
                    'symbol'  => 'L',
                    'decimal' => TRUE
                ],
                'MGA' => [
                    'label'   => 'Malagasy Ariary',
                    'format'  => 'before',
                    'locale'  => 'en_MG',
                    'symbol'  => 'Ar',
                    'decimal' => FALSE
                ],
                'MKD' => [
                    'label'   => 'Macedonian Denar',
                    'format'  => 'before',
                    'locale'  => 'mk_MK',
                    'symbol'  => '&#1076;&#1077;&#1085;',
                    'decimal' => TRUE
                ],
                'MNT' => [
                    'label'   => 'Mongolian Tögrög',
                    'format'  => 'before',
                    'locale'  => 'mn_Cyrl_MN',
                    'symbol'  => '&#8366;',
                    'decimal' => TRUE
                ],
                'MOP' => [
                    'label'   => 'Macanese Pataca',
                    'format'  => 'before',
                    'locale'  => 'pt_MO',
                    'symbol'  => 'MOP$',
                    'decimal' => TRUE
                ],
                'MRO' => [
                    'label'   => 'Mauritanian Ouguiya',
                    'format'  => 'before',
                    'locale'  => 'ar_MR',
                    'symbol'  => 'UM',
                    'decimal' => FALSE
                ],
                'MUR' => [
                    'label'   => 'Mauritian Rupee',
                    'format'  => 'before',
                    'locale'  => 'en_MU',
                    'symbol'  => '&#588;s',
                    'decimal' => TRUE
                ],
                'MVR' => [
                    'label'   => 'Maldivian Rufiyaa',
                    'format'  => 'before',
                    'locale'  => '',
                    'symbol'  => 'Rf',
                    'decimal' => TRUE
                ],
                'MWK' => [
                    'label'   => 'Malawian Kwacha',
                    'format'  => 'before',
                    'locale'  => 'en_MW',
                    'symbol'  => 'MK',
                    'decimal' => TRUE
                ],
                'MXN' => [
                    'label'   => 'Mexican Peso',
                    'format'  => 'before',
                    'locale'  => 'es_MX',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'MYR' => [
                    'label'   => 'Malaysian Ringgit',
                    'format'  => 'before',
                    'locale'  => 'ta_MY',
                    'symbol'  => 'RM',
                    'decimal' => TRUE
                ],
                'MZN' => [
                    'label'   => 'Mozambican Metical',
                    'format'  => 'before',
                    'locale'  => 'mgh_MZ',
                    'symbol'  => 'MT',
                    'decimal' => TRUE
                ],
                'NAD' => [
                    'label'   => 'Namibian Dollar',
                    'format'  => 'before',
                    'locale'  => 'naq_NA',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'NGN' => [
                    'label'   => 'Nigerian Naira',
                    'format'  => 'before',
                    'locale'  => 'en_NG',
                    'symbol'  => '&#8358;',
                    'decimal' => TRUE
                ],
                'NIO' => [
                    'label'   => 'Nicaraguan Córdoba',
                    'format'  => 'before',
                    'locale'  => 'es_NI',
                    'symbol'  => 'C$',
                    'decimal' => TRUE
                ],
                'NOK' => [
                    'label'   => 'Norwegian Krone',
                    'format'  => 'before',
                    'locale'  => 'se_NO',
                    'symbol'  => 'kr',
                    'decimal' => TRUE
                ],
                'NPR' => [
                    'label'   => 'Nepalese Rupee',
                    'format'  => 'before',
                    'locale'  => 'ne_NP',
                    'symbol'  => 'N&#588;s',
                    'decimal' => TRUE
                ],
                'NZD' => [
                    'label'   => 'New Zealand Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_NZ',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'PAB' => [
                    'label'   => 'Panamanian Balboa',
                    'format'  => 'before',
                    'locale'  => 'es_PA',
                    'symbol'  => 'B/.',
                    'decimal' => TRUE
                ],
                'PEN' => [
                    'label'   => 'Peruvian Nuevo Sol',
                    'format'  => 'before',
                    'locale'  => 'es_PE',
                    'symbol'  => 'S/.',
                    'decimal' => TRUE
                ],
                'PGK' => [
                    'label'   => 'Papua New Guinean Kina',
                    'format'  => 'before',
                    'locale'  => 'en_PG',
                    'symbol'  => 'K',
                    'decimal' => TRUE
                ],
                'PHP' => [
                    'label'   => 'Philippine Peso',
                    'format'  => 'before',
                    'locale'  => 'en_PH',
                    'symbol'  => '&#8369;',
                    'decimal' => TRUE
                ],
                'PKR' => [
                    'label'   => 'Pakistani Rupee',
                    'format'  => 'before',
                    'locale'  => 'en_PK',
                    'symbol'  => '&#588;s',
                    'decimal' => TRUE
                ],
                'PLN' => [
                    'label'   => 'Polish Złoty',
                    'format'  => 'after',
                    'locale'  => 'pl_PL',
                    'symbol'  => 'z&#322;',
                    'decimal' => TRUE
                ],
                'PYG' => [
                    'label'   => 'Paraguayan Guaraní',
                    'format'  => 'before',
                    'locale'  => 'es_PY',
                    'symbol'  => '&#8370;',
                    'decimal' => FALSE
                ],
                'QAR' => [
                    'label'   => 'Qatari Riyal',
                    'format'  => 'after',
                    'locale'  => 'ar_QA',
                    'symbol'  => '&#1585;.&#1602;',
                    'decimal' => TRUE
                ],
                'RON' => [
                    'label'   => 'Romanian Leu',
                    'format'  => 'before',
                    'locale'  => 'ro_RO',
                    'symbol'  => 'L',
                    'decimal' => TRUE
                ],
                'RSD' => [
                    'label'   => 'Serbian Dinar',
                    'format'  => 'before',
                    'locale'  => 'sr_Latn_RS',
                    'symbol'  => '&#1044;&#1080;&#1085;.',
                    'decimal' => TRUE
                ],
                'RUB' => [
                    'label'   => 'Russian Ruble',
                    'format'  => 'before',
                    'locale'  => 'ru_RU',
                    'symbol'  => '&#8381;',
                    'decimal' => TRUE
                ],
                'RWF' => [
                    'label'   => 'Rwandan Franc',
                    'format'  => 'before',
                    'locale'  => 'en_RW',
                    'symbol'  => 'RF',
                    'decimal' => FALSE
                ],
                'SAR' => [
                    'label'   => 'Saudi Riyal',
                    'format'  => 'after',
                    'locale'  => 'ar_SA',
                    'symbol'  => '&#1585;.&#1587;',
                    'decimal' => TRUE
                ],
                'SBD' => [
                    'label'   => 'Solomon Islands Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_SB',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'SCR' => [
                    'label'   => 'Seychellois Rupee',
                    'format'  => 'before',
                    'locale'  => 'fr_SC',
                    'symbol'  => '&#588;s',
                    'decimal' => TRUE
                ],
                'SEK' => [
                    'label'   => 'Swedish Krona',
                    'format'  => 'after',
                    'locale'  => 'sv_SE',
                    'symbol'  => 'kr',
                    'decimal' => TRUE
                ],
                'SGD' => [
                    'label'   => 'Singapore Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_SG',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'SHP' => [
                    'label'   => 'Saint Helenian Pound',
                    'format'  => 'before',
                    'locale'  => 'en_SH',
                    'symbol'  => '&pound;',
                    'decimal' => TRUE
                ],
                'SLL' => [
                    'label'   => 'Sierra Leonean Leone',
                    'format'  => 'before',
                    'locale'  => 'en_SL',
                    'symbol'  => 'Le',
                    'decimal' => TRUE
                ],
                'SOS' => [
                    'label'   => 'Somali Shilling',
                    'format'  => 'before',
                    'locale'  => 'so_SO',
                    'symbol'  => 'So. Sh.',
                    'decimal' => TRUE
                ],
                'SRD' => [
                    'label'   => 'Surinamese Dollar',
                    'format'  => 'before',
                    'locale'  => 'nl_SR',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'STD' => [
                    'label'   => 'São Tomé and Príncipe Dobra',
                    'format'  => 'before',
                    'locale'  => 'pt_ST',
                    'symbol'  => 'Db',
                    'decimal' => TRUE
                ],
                'SZL' => [
                    'label'   => 'Swazi Lilangeni',
                    'format'  => 'before',
                    'locale'  => 'en_SZ',
                    'symbol'  => 'L',
                    'decimal' => TRUE
                ],
                'THB' => [
                    'label'   => 'Thai Baht',
                    'format'  => 'before',
                    'locale'  => 'th_TH',
                    'symbol'  => '&#3647;',
                    'decimal' => TRUE
                ],
                'TJS' => [
                    'label'   => 'Tajikistani Somoni',
                    'format'  => 'before',
                    'locale'  => 'tg_Cyrl_TJ',
                    'symbol'  => 'SM',
                    'decimal' => TRUE
                ],
                'TOP' => [
                    'label'   => 'Tongan Paʻanga',
                    'format'  => 'before',
                    'locale'  => 'en_TO',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'TRY' => [
                    'label'   => 'Turkish Lira',
                    'format'  => 'before',
                    'locale'  => 'tr_TR',
                    'symbol'  => '&#8378;',
                    'decimal' => TRUE
                ],
                'TTD' => [
                    'label'   => 'Trinidad and Tobago Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_TT',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'TWD' => [
                    'label'   => 'New Taiwan Dollar',
                    'format'  => 'before',
                    'locale'  => 'zh_Hant_TW',
                    'symbol'  => 'NT$',
                    'decimal' => FALSE
                ],
                'TZS' => [
                    'label'   => 'Tanzanian Shilling',
                    'format'  => 'before',
                    'locale'  => 'en_TZ',
                    'symbol'  => 'TSh',
                    'decimal' => TRUE
                ],
                'UAH' => [
                    'label'   => 'Ukrainian Hryvnia',
                    'format'  => 'before',
                    'locale'  => 'uk_UA',
                    'symbol'  => '&#8372;',
                    'decimal' => TRUE
                ],
                'UGX' => [
                    'label'   => 'Ugandan Shilling',
                    'format'  => 'before',
                    'locale'  => 'en_UG',
                    'symbol'  => 'USh',
                    'decimal' => FALSE
                ],
                'USD' => [
                    'label'   => 'United States Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_US',
                    'symbol'  => '$',
                    'decimal' => TRUE
                ],
                'UYU' => [
                    'label'   => 'Uruguayan Peso',
                    'format'  => 'before',
                    'locale'  => 'es_UY',
                    'symbol'  => '$U',
                    'decimal' => TRUE
                ],
                'UZS' => [
                    'label'   => 'Uzbekistani Som',
                    'format'  => 'before',
                    'locale'  => 'uz_Latn_UZ',
                    'symbol'  => '&#1083;&#1074;',
                    'decimal' => TRUE
                ],
                'VEF' => [
                    'label'   => 'Venezuelan Bolívar',
                    'format'  => 'before',
                    'locale'  => 'es_VE',
                    'symbol'  => 'Bs',
                    'decimal' => TRUE
                ],
                'VND' => [
                    'label'   => 'Vietnamese Đồng',
                    'format'  => 'after',
                    'locale'  => 'vi_VN',
                    'symbol'  => '&#8363;',
                    'decimal' => FALSE
                ],
                'VUV' => [
                    'label'   => 'Vanuatu Vatu',
                    'format'  => 'after',
                    'locale'  => 'en_VU',
                    'symbol'  => 'VT',
                    'decimal' => FALSE
                ],
                'WST' => [
                    'label'   => 'Samoan Tala',
                    'format'  => 'before',
                    'locale'  => 'en_WS',
                    'symbol'  => 'WS$',
                    'decimal' => TRUE
                ],
                'XAF' => [
                    'label'   => 'Central African Cfa Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_CF',
                    'symbol'  => 'CFA',
                    'decimal' => FALSE
                ],
                'XCD' => [
                    'label'   => 'East Caribbean Dollar',
                    'format'  => 'before',
                    'locale'  => 'en_AI',
                    'symbol'  => 'EC$',
                    'decimal' => TRUE
                ],
                'XOF' => [
                    'label'   => 'West African Cfa Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_BF',
                    'symbol'  => 'CFA',
                    'decimal' => FALSE
                ],
                'XPF' => [
                    'label'   => 'Cfp Franc',
                    'format'  => 'before',
                    'locale'  => 'fr_PF',
                    'symbol'  => 'F',
                    'decimal' => FALSE
                ],
                'YER' => [
                    'label'   => 'Yemeni Rial',
                    'format'  => 'after',
                    'locale'  => 'ar_YE',
                    'symbol'  => '&#65020;',
                    'decimal' => TRUE
                ],
                'ZAR' => [
                    'label'   => 'South African Rand',
                    'format'  => 'before',
                    'locale'  => 'en_LS',
                    'symbol'  => 'R',
                    'decimal' => TRUE
                ],
                'ZMW' => [
                    'label'   => 'Zambian Kwacha',
                    'format'  => 'before',
                    'locale'  => 'en_ZM',
                    'symbol'  => 'ZMW',
                    'decimal' => TRUE
                ],
            ];
            if (NULL === $code) {
                return $currencies;
            }

            return $currencies[ $code ];
        }

        /**
         * @param int $from
         * @param int $to
         *
         * @return string
         */
        public static function human_time_diff(int $from, int $to = 0): string
        {
            if (empty($to)) {
                $to = time();
            }

            $diff = (int)abs($to - $from);

            if ($diff < MINUTE_IN_SECONDS) {
                $secs = $diff;
                if ($secs <= 1) {
                    $secs = 1;
                }
                $since = sprintf(
                /* translators: %d: number of seconds */
                    _n('%d second', '%d seconds', $secs),
                    $secs
                );
            } elseif ($diff < HOUR_IN_SECONDS && $diff >= MINUTE_IN_SECONDS) {
                $mins  = round($diff / MINUTE_IN_SECONDS);
                $since = sprintf(
                /* translators: %d: number of minutes */
                    _n('%d m', '%d m', $mins),
                    $mins
                );
            } elseif ($diff < DAY_IN_SECONDS && $diff >= HOUR_IN_SECONDS) {
                $hours = round($diff / HOUR_IN_SECONDS);
                $diff  -= $hours * HOUR_IN_SECONDS;
                $mins  = round($diff / MINUTE_IN_SECONDS);
                $since = sprintf(
                /* translators: %d: number of hours */
                    _n('%d h', '%d h', $hours),
                    $hours
                );
                if ($mins > 0) {
                    $since .= ' ' . sprintf(
                        /* translators: %d: number of minutes */
                            _n('%d m', '%d m', $mins),
                            $mins
                        );
                }
            } elseif ($diff < WEEK_IN_SECONDS && $diff >= DAY_IN_SECONDS) {
                $days = round($diff / DAY_IN_SECONDS);
                if ($days <= 1) {
                    $days = 1;
                }
                $since = sprintf(
                /* translators: %d: number of days */
                    _n('%d day', '%d days', $days),
                    $days
                );
            } elseif ($diff < MONTH_IN_SECONDS && $diff >= WEEK_IN_SECONDS) {
                $weeks = round($diff / WEEK_IN_SECONDS);
                if ($weeks <= 1) {
                    $weeks = 1;
                }
                $since = sprintf(
                /* translators: %d: number of weeks */
                    _n('%d week', '%d weeks', $weeks),
                    $weeks
                );
            } elseif ($diff < YEAR_IN_SECONDS && $diff >= MONTH_IN_SECONDS) {
                $months = round($diff / MONTH_IN_SECONDS);
                if ($months <= 1) {
                    $months = 1;
                }
                $since = sprintf(
                /* translators: %d: number of months */
                    _n('%d month', '%d months', $months),
                    $months
                );
            } elseif ($diff >= YEAR_IN_SECONDS) {
                $years = round($diff / YEAR_IN_SECONDS);
                if ($years <= 1) {
                    $years = 1;
                }
                $since = sprintf(
                /* translators: %d: number of years */
                    _n('%d year', '%d years', $years),
                    $years
                );
            }

            return $since;
        }

        public static function get_request_headers()
        {
            if (!function_exists('getallheaders')) {
                $headers = [];
                foreach ($_SERVER as $name => $value) {
                    if (strpos($name, 'HTTP_') === 0) {
                        $headers[ str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5))))) ] = $value;
                    }
                }

                return $headers;
            }

            return getallheaders();
        }

    }
}