<?php

namespace VSHM;

use VSHM\Bus\UpdateProviderProperty;
use VSHM\Modules\Gcal2Ways;
use VSHM\Providers\Promotions;
use VSHM\Providers\Reservations;
use VSHM\Providers\ReservationsData;
use VSHM\Providers\ServiceProviderCustomData;
use VSHM\Providers\ServiceProviders;
use VSHM\Providers\Services;
use VSHM\Providers\ServicesData;
use VSHM\Settings\GoogleAllowSlotCommands;
use VSHM\Settings\GoogleApiApplicationName;
use VSHM\Settings\GoogleApiClientId;
use VSHM\Settings\GoogleApiClientSecret;
use VSHM\Settings\Provider\AllowedServices;
use VSHM\Settings\Provider\GoogleApiToken;
use VSHM\Settings\ZoomApiKey;
use VSHM\Settings\ZoomApiSecret;

defined('ABSPATH') || exit;

/**
 * Class Update
 *
 * @package VSHM
 * @author  VonStroheim
 */
class Update
{

    public static function get_stored_version()
    {
        return get_option('tbk_version');
    }

    public static function is_migrated_from_v2()
    {
        return get_option('tbk_migrated_from_v2');
    }

    public static function is_v2_data_present(): bool
    {
        return (bool)get_option('team_booking');
    }

    public static function is_awaiting_migration_from_v2(): bool
    {
        return self::is_v2_data_present() && !self::is_migrated_from_v2();
    }

    public static function update_stored_version(): void
    {
        update_option('tbk_version', vshm()->plugin['VERSION']);
    }

    public static function set_migrated_from_v2(bool $migrated = TRUE): void
    {
        update_option('tbk_migrated_from_v2', $migrated);
    }

    public static function remove_v2_data()
    {
        DB::drop_table('teambooking_events');
        DB::drop_table('teambooking_promotions');
        DB::drop_table('teambooking_reservations');
        DB::drop_table('teambooking_sessions');
        wp_clear_scheduled_hook('tb-db-cleaning-routine');
        wp_clear_scheduled_hook('tb_email_reminder_handler');
        $post_args = [
            'post_type' => ['tbk_service', 'tbk_form'],
            'nopaging'  => TRUE
        ];
        $posts     = get_posts($post_args);
        foreach ($posts as $post) {
            $properties = array_keys(get_post_custom($post->ID));
            foreach ($properties as $property) {
                delete_post_meta_by_key($property);
            }
            wp_delete_post($post->ID, TRUE);
        }
        delete_option('tmbkg_zoom');
        delete_option('team_booking');
    }

    public static function maybe_update(): void
    {
        if (vshm()->plugin['VERSION'] !== self::get_stored_version()) {
            self::update_stored_version();
        }
    }

    /**
     * @return bool
     */
    public static function to_3(): bool
    {
        /**
         * Plugin settings
         */
        $settings      = TbkSettings::translateSettings();
        $base_settings = [];
        $gcal_settings = [];
        foreach ($settings as $key => $setting) {
            if ($key === GoogleApiClientId::ID) {
                $gcal_settings[ $key ] = $setting;
                continue;
            }
            if ($key === GoogleApiClientSecret::ID) {
                $gcal_settings[ $key ] = $setting;
                continue;
            }
            if ($key === GoogleApiApplicationName::ID) {
                $gcal_settings[ $key ] = $setting;
                continue;
            }
            if ($key === GoogleAllowSlotCommands::ID) {
                $gcal_settings[ $key ] = $setting;
                continue;
            }
            $base_settings[ $key ] = $setting;
        }

        $zoomData = get_option('tmbkg_zoom');
        if (is_array($zoomData)) {
            foreach ($zoomData as $key => $zoomDatum) {
                if ($key === 'zoom_apiKey') {
                    $base_settings[ ZoomApiKey::ID ] = $zoomDatum;
                }
                if ($key === 'zoom_apiSecret') {
                    $base_settings[ ZoomApiSecret::ID ] = $zoomDatum;
                }
            }

        }

        vshm()->settings();

        update_option(vshm()->get_settings_name(), $base_settings);
        update_option(Gcal2Ways::OPTIONS_TAG, $gcal_settings);

        /**
         * Services
         */
        $services = TbkSettings::translateServices();

        $existingServices = array_column(Services::provide(), 'name', 'id');

        Services::storeMany($services);

        $data_to_store = [];
        foreach ($services as $service) {
            if (isset($service['data']) && is_array($service['data'])) {
                foreach ($service['data'] as $key => $value) {
                    $data_to_store[] = [
                        'service_id' => $service['id'],
                        'key'        => $key,
                        'value'      => $value
                    ];
                }
            }
        }

        ServicesData::storeMany($data_to_store);

        /**
         * Forms
         */
        TbkSettings::translateForms();

        /**
         * Custom service settings
         */
        $settings = TbkSettings::translateCustomServiceSettings();

        ServiceProviderCustomData::storeMany($settings);

        // Filling gaps with defaults
        $services  = Services::provide();
        $providers = ServiceProviders::provide();

        $to_insert = [];
        foreach ($services as $service) {
            if ($service['class'] === 'unscheduled') {
                continue;
            }
            $defaults = apply_filters('vshm_default_service_personal_settings', [], vshm()->plugin['SLUG'], $service);
            foreach ($providers as $provider) {
                foreach ($defaults as $key => $default) {
                    $to_insert[] = [
                        'service_id'  => $service['id'],
                        'provider_id' => $provider['id'],
                        'key'         => $key,
                        'value'       => $default,
                    ];
                }
            }
        }
        ServiceProviderCustomData::storeMany($to_insert);

        /**
         * Reservations
         */
        TbkSettings::translateReservations();

        /**
         * Promotions
         */
        $promotions = TbkSettings::translatePromotions();
        if (!empty($promotions)) {
            Promotions::storeMany($promotions);
        }

        self::set_migrated_from_v2(TRUE);

        return TRUE;
    }

}